///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Copyright Nurve Networks LLC 2009
// 
// Filename: CHAM_AVR_KEYBOARD_DEMO_01.C
//
// Original Author: Andre' LaMothe
// 
// Last Modified: 10.18.09
//
// Description: Read local PS/2 keyboard connected to Chameleon prints the keys as user presses them.
//                
// Overview: Prints local PS/2 keyboard key presses to both NTSC and VGA screens. Make sure to have keyboard plugged in
//           and hit reset to start program.
//
// Chameleon PIC version nearly identical
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
 

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// __AVR_ATmega328P__ , __AVR_ATmega168__
#define __AVR_ATmega328P__ 

// include everything 
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <util/delay.h> 
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <compat/twi.h>
#include <avr/wdt.h>

// include headers required for demo
#include "CHAM_AVR_SYSTEM_V010.h"             // you need this one always
#include "CHAM_AVR_TWI_SPI_DRV_V010.h"        // you need this always
#include "CHAM_AVR_NTSC_DRV_V010.h"           // needed for NTSC driver
#include "CHAM_AVR_VGA_DRV_V010.h"            // needed for VGA driver
#include "CHAM_AVR_KEYBOARD_DRV_V010.h"       // needed for keyboard driver

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES AND MACROS /////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// define CPU frequency in Mhz here if not defined in Makefile or other includes, compiler will throw a warning, ignore
#ifndef F_CPU
#define F_CPU 16000000UL // 28636360UL, 14318180UL, 21477270 UL
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GLOBALS ////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// a little fun
char access_code[] = "morpheous";
int access_code_index = 0;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MAIN ///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int main(void)
{
int key;          // key that keyboard returns
char sbuffer[64]; // printing buffer

// initialize and set SPI rate
SPI_Init( SPI_DEFAULT_RATE );

// give Propeller driver time to boot
_delay_ms( 2500 );

// clear screens
NTSC_ClearScreen();
VGA_ClearScreen();

// set NTSC color
NTSC_Color(0);
NTSC_Term_Print(" PS/2 Keyboard Demo");
NTSC_Term_Char( 0x0D );
NTSC_Term_Print(" Press keys on local keyboard.");
NTSC_Term_Char( 0x0D );

// set VGA color
VGA_Color(0);
VGA_Term_Char( 0x0D );
VGA_Term_Print(" Press keys on local keyboard.");
VGA_Term_Char( 0x0D );

// load keyboard driver (this driver is already loaded by default, but this call can't hurt)
Keyboard_Load();

// enter infinite loop...
while(1)
     {
     // read next key, return 0 (null if no key), wait for key press
     while ( (key = Keyboard_Key()) == 0);
  
     // build display string
     if (key >= 32)
        sprintf(sbuffer, "Key Code = %d [%c]                ", (int)key, (char)key);
     else
        sprintf(sbuffer, "Key Code = %d [non-printable]     ", (int)key);

     // position cursor
     NTSC_SetXY(1, 4);

     // print on NTSC terminal screen 
     NTSC_Term_Print(sbuffer);
     NTSC_Term_Char( 0x0D );

     // position cursor
     VGA_SetXY(1, 4);

     // print on VGA terminal screen 
     VGA_Term_Print(sbuffer);
     VGA_Term_Char( 0x0D );

     //////////////////////////////////////////////////////////////////////////////////       
     // little string compare state machine for fun, when user types in access code
     // it displays a little message at bottom
     if (key == access_code[ access_code_index ])
        {
        // consume character, test for match?
        if (++access_code_index == strlen(access_code))
           {
           NTSC_SetXY(1, 22);
           NTSC_Term_Print("Welcome NEO...Access Granted.");
           
           VGA_SetXY(1, 14);
           VGA_Term_Print("Welcome NEO...Access Granted.");
         
           access_code_index = 0;

           } // end if match

        } // end end if
     else // reset access code, start over
       access_code_index = 0;
     //////////////////////////////////////////////////////////////////////////////////     

     // slow things down a bit, so we can read the text!
     _delay_ms(10);

     } // end while    

} // end main
