///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Copyright Nurve Networks LLC 2009
// 
// Filename: CHAM_AVR_MOUSE_DEMO_01.C
//
// Original Author: Andre' LaMothe
// 
// Last Modified: 10.18.09
//
// Description:  Uses mouse to draw on screen(s).
//  
// Overview: Uses mouse to draw on NTSC and VGA screen at the same time.
// Chameleon PIC version nearly identical
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
 

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// __AVR_ATmega328P__ , __AVR_ATmega168__
#define __AVR_ATmega328P__ 

// include everything 
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <util/delay.h> 
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <compat/twi.h>
#include <avr/wdt.h>

// include headers required for demo
#include "CHAM_AVR_SYSTEM_V010.h"             // you need this one always
#include "CHAM_AVR_TWI_SPI_DRV_V010.h"        // you need this always
#include "CHAM_AVR_NTSC_DRV_V010.h"           // needed for NTSC driver
#include "CHAM_AVR_VGA_DRV_V010.h"            // needed for VGA driver
#include "CHAM_AVR_MOUSE_DRV_V010.h"          // needed for mouse driver

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES AND MACROS /////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// define CPU frequency in Mhz here if not defined in Makefile or other includes, compiler will throw a warning, ignore
#ifndef F_CPU
#define F_CPU 16000000UL // 28636360UL, 14318180UL, 21477270 UL
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GLOBALS/////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MAIN ///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int main(void)
{
int x=16, y=12;             // position of cursor
unsigned char cursor = 'X'; // start pointer looking like 'X'
gid_event event;            // generic "event" for mouse, etc.

/*
// generic input device data record type, helps get back data from various multidata calls
typedef struct gid_event_type
    {

    int x,y,z;    // position data
    int buttons;  // buttons, bit encoded
                  // bit4 = right-side button
                  // bit3 = left-side button
                  // bit2 = center/scrollwheel button
                  // bit1 = right button
                  // bit0 = left button

    } gid_event, *gid_event_ptr;
*/

// initialize and set SPI rate
SPI_Init( SPI_DEFAULT_RATE );

// allow Propeller driver time to boot, this is necessary before sending commands, assuming a system reset at the same time
_delay_ms( 2500 );

// load the mouse
Mouse_Load();

// clear screens and display instructions
NTSC_ClearScreen();
NTSC_Color(0);
NTSC_Term_Print(" PS/2 Mouse Test - ChamPaint");
NTSC_Term_Char( 0x0D );
NTSC_Term_Char( 0x0D );
NTSC_Term_Print(" - Left button to clear screen.");
NTSC_Term_Char( 0x0D );
NTSC_Term_Print(" - Right to change cursor.");
NTSC_Term_Char( 0x0D );

VGA_ClearScreen();
VGA_Color(0);
VGA_Term_Print(" PS/2 Mouse Test - ChamPaint");
VGA_Term_Char( 0x0D );
VGA_Term_Char( 0x0D );
VGA_Term_Print(" - Left button to clear screen.");
VGA_Term_Char( 0x0D );
VGA_Term_Print(" - Right to change cursor.");
VGA_Term_Char( 0x0D );

// main event loop
while(1)
   {
   // read the mouse event (if one)
   Mouse_Read(&event); 

   // compute position
   event.x = event.x / 10;
   event.y = event.y / 10;

   // left button hit? Clear screen?
   if (event.buttons == 0X01)
     {
     // clear the surface
     NTSC_ClearScreen();
     VGA_ClearScreen();
     } // end if

   // right button hit? update cursor character?
   if (event.buttons == 0X02)
     {
     // update the cursor character
     if (++cursor >= (32+144))
       cursor = 33;

     } // end if

   // motion clamping                  
   if (event.x < -2)
     event.x = -2;
   else 
   if (event.x > 2)
       event.x = 2;

   if (event.y < -2)
     event.y = -2;
   else 
   if (event.y > 2)
       event.y = 2;

   // update position
   x += event.x;
   y -= event.y;

   // draw on both screens
   NTSC_SetXY( x, y );
   NTSC_Term_Char( cursor );

   VGA_SetXY( x, y/2 );
   VGA_Term_Char( cursor );

   // slow things down a bit
   _delay_ms( 5 );
   } // end while

} // end main
