///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Copyright Nurve Networks LLC 2009
// 
// Filename: CHAM_AVR_NTSC_GLOW_DEMO_01.c
//
// Original Author: Andre' LaMothe
// 
// Last Modified: 10.18.09
//
// Description: NTSC demo that "glows" the top and bottom overscan region using some of the NTSC specific 
// functionality in the Default2 version of the Propeller driver.
//  
// Overview: Simply prints out some text then glows the top and bottom overscan, by slowing cycling thru all the colors
// in the palette.
//
// Chameleon PIC version nearly identical
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
 

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// __AVR_ATmega328P__ , __AVR_ATmega168__
#define __AVR_ATmega328P__ 

// include everything 
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <util/delay.h> 
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <compat/twi.h>
#include <avr/wdt.h>

// include headers required for demo
#include "CHAM_AVR_SYSTEM_V010.h"             // you need this one always
#include "CHAM_AVR_TWI_SPI_DRV_V010.h"        // you need this always
#include "CHAM_AVR_NTSC_DRV_V010.h"           // needed for NTSC driver
#include "CHAM_AVR_GFX_DRV_V010.h"            // needed for GFX driver

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES AND MACROS /////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// define CPU frequency in Mhz here if not defined in Makefile or other includes, compiler will throw a warning, ignore
#ifndef F_CPU
#define F_CPU 16000000UL // 28636360UL, 14318180UL, 21477270 UL
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MAIN ///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int main(void)
{
unsigned char color, bright, col;

// initialize and set SPI rate
SPI_Init( SPI_DEFAULT_RATE );

// give Prop a moment before sending it commands, the boot process is about a second
// if you like you can speed it up, by removing the Prop driver's "LED blink" sequence in the driver
// or speed it up, but commands that are sent before the Prop driver is done booting will be ignored.
// therefore, if you have a lot of set up work to do, then you don't require a delay, but if you jump right
// into commands, then you need a good 1.5 - 2 second delay
_delay_ms( 2500 ); 

// clear screens
NTSC_ClearScreen();

// print on NTSC terminal screen 
NTSC_Color(0);

// print text
NTSC_Term_Print("Cylone Base Star to Outpost...");
NTSC_Term_Char( 0x0D ); // newline
NTSC_Term_Char( 0x0D ); // newline
_delay_ms( 1000 ); 

NTSC_Term_Print("Message is:");
NTSC_Term_Char( 0x0D ); // newline
NTSC_Term_Char( 0x0D ); // newline
_delay_ms( 2000 ); 

NTSC_Term_Print("Destroy Caprica!");
NTSC_Term_Char( 0x0D ); // newline
_delay_ms( 500 ); 

NTSC_Term_Print("Destroy Caprica!");
NTSC_Term_Char( 0x0D ); // newline
_delay_ms( 500 ); 

NTSC_Term_Print("Destroy Caprica!");
NTSC_Term_Char( 0x0D ); // newline
_delay_ms( 1500 ); 

// position the cursor and print in center
NTSC_SetXY(8, 12);
NTSC_Term_Print("BY YOUR COMMAND...");

// enter infinite loop...glowing the top and bottom overscan, tile engine functions required for this
while(1)
     {
     // glow thru every color, every brightness
     for (col = 0; col <= 15 ; col++)
         {
         for (bright = 2; bright < 6; bright++)
            {

            // build up Propeller compatible color [chroma:4 | chroma bit:1 | luma:3]
            color = (col << 4) | (0b00001000) | (bright);
            
            // set top overscan
            GFX_Set_Top_Overscan_Color( color & 0xFF );

            // set bottom overscan
            GFX_Set_Bottom_Overscan_Color( color  & 0xFF );

            // slow things down a bit, so we can read the text!
            _delay_ms(100);

            } // end for bright

       } // end for color

     } // end while    

} // end main
