///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Copyright Nurve Networks LLC 2009
// 
// Filename: CHAM_AVR_NTSC_SMOOTH_SCROLL_DEMO_01.c
//
// Original Author: Andre' LaMothe
// 
// Last Modified: 10.18.09
//
// Description: NTSC demo of smooth scrolling with the advaced NTSC tile engine.
//  
// Overview: Simply smooth scrolls some tiles on the screen up and down with the fine scroll register.
// Chameleon PIC version nearly identical
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
 

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// __AVR_ATmega328P__ , __AVR_ATmega168__
#define __AVR_ATmega328P__ 

// include everything 
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <util/delay.h> 
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <compat/twi.h>
#include <avr/wdt.h>

// include headers required for demo
#include "CHAM_AVR_SYSTEM_V010.h"             // you need this one always
#include "CHAM_AVR_TWI_SPI_DRV_V010.h"        // you need this always
#include "CHAM_AVR_NTSC_DRV_V010.h"           // needed for NTSC driver
#include "CHAM_AVR_GFX_DRV_V010.h"            // needed for special GFX commands like fine scrolling

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES AND MACROS /////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// define CPU frequency in Mhz here if not defined in Makefile or other includes, compiler will throw a warning, ignore
#ifndef F_CPU
#define F_CPU 16000000UL // 28636360UL, 14318180UL, 21477270 UL
#endif

#define SPACE_INVADERS_CHAR_OFFSET (32+4*32+13) // 32 is where "blank/space" is mapped, so our character set starts from 32...
#define SPACE_INVADERS_ORIGIN_X    6
#define SPACE_INVADERS_ORIGIN_Y    8 
#define NUM_SPACE_INVADERS_PER_ROW 10

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MAIN ///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int main(void)
{
int x,y, scroll_y = 0, scroll_dy = 1;

// initialize and set SPI rate
SPI_Init( SPI_DEFAULT_RATE );

// give Prop a moment before sending it commands, the boot process is about a second
// if you like you can speed it up, by removing the Prop driver's "LED blink" sequence in the driver
// or speed it up, but commands that are sent before the Prop driver is done booting will be ignored.
// therefore, if you have a lot of set up work to do, then you don't require a delay, but if you jump right
// into commands, then you need a good 1.5 - 2 second delay
_delay_ms( 2500 );

// clear screens
NTSC_ClearScreen();

// basic white
NTSC_Color(0);

// draw a matrix of space invaders with extended characters 
for (y = 0; y < 6; y++)
    {
    for (x = 0; x < NUM_SPACE_INVADERS_PER_ROW; x++)
        {
        // position next invader
        NTSC_SetXY(x*2 + SPACE_INVADERS_ORIGIN_X , y*2 + SPACE_INVADERS_ORIGIN_Y);

        // draw next invader
        NTSC_Term_Char( SPACE_INVADERS_CHAR_OFFSET + (y % 3) );
        } // end for x
    } // end for y

// enter infinite loop...
while(1)
     {
     // set smooth scroll vertical position register 0..7
     GFX_Set_VScroll( scroll_y );

     // change scroll position
     scroll_y += scroll_dy;

     // test for limit
     if ( (scroll_y > 7) || (scroll_y < 0) )
        {
        scroll_dy = -scroll_dy;
        scroll_y += scroll_dy;
        } // end if

     // slow things down a bit, so we can read the text!
     _delay_ms(50);

     } // end while    

} // end main
