///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Copyright Nurve Networks LLC 2009
// 
// Filename: CHAM_AVR_SOUND_DRV_V010.C
//
// Original Author: Andre' LaMothe
// 
// Last Modified: 8.28.09
//
// Description: Sounds driver API.
//  
// Overview:
//
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// this processor include needs to happen in the master file
// #define __AVR_ATmega168__

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <compat/twi.h>
#include <avr/pgmspace.h>

// include local Chameleon AVR API header files
#include "CHAM_AVR_SYSTEM_V010.h"
#include "CHAM_AVR_TWI_SPI_DRV_V010.h"
#include "CHAM_AVR_SOUND_DRV_V010.h"

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MACROS /////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES/////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TYPES/CLASSES //////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GLOBALS ////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// EXTERNALS //////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PROTOTYPES /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FUNCTIONS  /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Sound driver support
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int Sound_Play(int channel, int frequency, int duration)
{
// plays a sound on the speaker using PWM, channel, frequency, duration

int low_data, hi_data; 

// extract lower 2-bits of channel, shift to upper 2-bits of byte, then extract lower 3-bits of duration,
// shift behind channel, then finally OR in upper 3-bits of frequency
hi_data = (channel << 6) | (duration << 3) | (frequency >> 8);

// mask off upper 6-bits of frequency and send only lower 8-bits 
low_data = (frequency & 0xFF);

// final data format
// arg4 [cc.dddfff ] | arg5 [ ffffffff ]
// channel 2-bit (0..3) | duration 3-bit 0..7 | frequency 11-bit (0 turn off channel)
SPI_Prop_Send_Cmd( SND_CMD_PLAYSOUNDFM, low_data, hi_data);
//_delay_us(SPI_PROP_DELAY_LONG_US);

// return success
return(1);

} // end Sound_Play

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int Sound_Stop(int channel)
{
// stops a sound channel 0..3
SPI_Prop_Send_Cmd( SND_CMD_STOPSOUND, channel, 0);
//_delay_us(SPI_PROP_DELAY_LONG_US);

// return success
return(1);

} // end  Sound_Stop

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int Sound_StopAll(void)
{
// stops all sound channels
SPI_Prop_Send_Cmd( SND_CMD_STOPALLSOUNDS, 0, 0);
//_delay_us(SPI_PROP_DELAY_LONG_US);

// return success
return(1);

} // end Sound_StopAll
 
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int Sound_Update(int channel, int frequency, int volume)
{
// allows you to update a sound channel's volume or frequency while its playing

int data_low, data_high;

// sets the frequency of a playing sound channel
// data8[cc.xxx.fff ] | data8[ ffffffff ]
// channel 2-bit (0..3) | dummy 3-bit xxx| frequency 11-bit (0 turn off channel)

// prepare data in bit encoded format
data_high = (channel << 6) | (frequency >> 8); 
data_low  = (frequency & 0xFF); 

// send the command to update frequency
SPI_Prop_Send_Cmd( SND_CMD_SETFREQ, data_low, data_high );
//_delay_us(SPI_PROP_DELAY_SHORT_US);

// sets the volume of the playing sound channel
// data8[cc.xxxxxx ] | data8[ vvvvvvvv ]
// channel 2-bit (0..3) | dummy 6-bit xxxxxx| volume 8-bit (1..255, 0 leave volume same)

data_high = (channel << 6);
data_low  = (volume & 0xFF); 

// send command to update volume
SPI_Prop_Send_Cmd( SND_CMD_SETVOLUME, data_low, data_high);
//_delay_us(SPI_PROP_DELAY_LONG_US);

// return success
return(1);

} // end Sound_Update

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
