///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Copyright Nurve Networks LLC 2009
// 
// Filename: CHAM_AVR_STATUS_LED_DEMO_01.c
//
// Original Author: Andre' LaMothe
// 
// Last Modified: 10.20.09
//
// Description: Demo of controlling the status LED with the modified Propeller Driver
//  
// Overview: In the manual there is a section that covers how to modify the Propeller master control program/driver
// to add features and commands, section "oping Your Own Propeller Drivers". As an example, we go into the propeller 
// driver and add two commands:
//
//  ' added status LED commands added for example in manual "Developing Your Own Propeller Drivers". 
//  STATUS_LED_ON                = 100
//  STATUS_LED_OFF               = 101
//
// and then this little chunk of code in the main message dispatcher:
//
// ' // Propeller status LED commands book example on how to modify the driver 
// STATUS_LED_ON:
//  OUTA[25] := 1
//        
// STATUS_LED_OFF:
//  OUTA[25] := 0
// ' // end added code for status LED control
//
// This C program then uses these new commands to control the Status LED by sending SPI commands directly.
//
// Chameleon PIC version of program nearly identical
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
 

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// __AVR_ATmega328P__ , __AVR_ATmega168__
#define __AVR_ATmega328P__ 

// include everything 
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <util/delay.h> 
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <compat/twi.h>
#include <avr/wdt.h>

// include headers required for demo
#include "CHAM_AVR_SYSTEM_V010.h"             // you need this one always
#include "CHAM_AVR_TWI_SPI_DRV_V010.h"        // you need this always
#include "CHAM_AVR_NTSC_DRV_V010.h"           // needed for NTSC driver

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES AND MACROS /////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// define CPU frequency in Mhz here if not defined in Makefile or other includes, compiler will throw a warning, ignore
#ifndef F_CPU
#define F_CPU 16000000UL // 28636360UL, 14318180UL, 21477270 UL
#endif

// new commands for status LED control
#define STATUS_LED_ON  100
#define STATUS_LED_OFF 101

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MAIN ///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int main(void)
{
// initialize and set SPI rate
SPI_Init( SPI_DEFAULT_RATE );

// give Prop a moment before sending it commands, the boot process is about a second
// if you like you can speed it up, by removing the Prop driver's "LED blink" sequence in the driver
// or speed it up, but commands that are sent before the Prop driver is done booting will be ignored.
// therefore, if you have a lot of set up work to do, then you don't require a delay, but if you jump right
// into commands, then you need a good 1.5 - 2 second delay
_delay_ms( 2500 );

// clear screens
NTSC_ClearScreen();
NTSC_Color(0);

// enter infinite loop...
while(1)
     {
     // print on NTSC terminal screen 
     NTSC_Term_Print(" Status LED ON");
     NTSC_Term_Char( 0x0D );

     // send command to turn on LED
     SPI_Prop_Send_Cmd(STATUS_LED_ON, 0,0 );

     // slow things down a bit, so we can read the text!
     _delay_ms(250);
     
     NTSC_Term_Print(" Status LED OFF");
     NTSC_Term_Char( 0x0D );

     // send command to turn off LED
     SPI_Prop_Send_Cmd(STATUS_LED_OFF, 0,0 );

     // slow things down a bit, so we can read the text!
     _delay_ms(250);

     } // end while    

} // end main
