///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Copyright Nurve Networks LLC 2009
// 
// Filename: CHAMELEON_UART_DRV_V010.h
//
// Original Author: Andre' LaMothe
// 
// Last Modified: 4.28.08
//
// Description: UART library file header
//
// Overview:
// 
// I/O pin map for UART interface
//  
// SER_TX | PD1 | Pin 3 | Output
// SER_RX | PD0 | Pin 4 | Input
//
// Note: The UART can also be used as another SPI channel.
// 
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// watch for multiple inclusions
#ifndef CHAM_AVR_UART_DRV
#define CHAM_AVR_UART_DRV

 // support C++ compilers
 #ifdef __cplusplus
 extern "C" {
 #endif 

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MACROS /////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES/////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* default baud rate */
#define UART_DEFAULT_BAUDRATE     115200 // slow down to 57600 if you have trouble

// transmit and receive buffer constants
#define UART_TX_BUFF_SIZE          32
#define UART_RX_BUFF_SIZE          32

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TYPES/CLASSES //////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GLOBALS ////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// global that holds UART current baud rate
extern long g_baudrate;

// receiver and transmitter head and tail indices for circular buffers
extern int g_uart_rx_head;
extern int g_uart_rx_tail;
extern int g_uart_tx_head;
extern int g_uart_tx_tail;

// buffers for uart interrupt driver
extern UCHAR g_uart_buffer_rx[UART_RX_BUFF_SIZE];    // receiver buffer for incoming characters
extern UCHAR g_uart_buffer_tx[UART_RX_BUFF_SIZE];    // transmit buffer for outgoing characters

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// EXTERNALS //////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PROTOTYPES /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// initializes UART, ISRs, etc.
int UART_Init(long baudrate);

// shuts down the UART
int UART_Shutdown(void);

// transmits a single character immediately
void UART_Transmit(unsigned char ch);

// transmits a character using interrupt buffer
int UART_putc(unsigned char ch);

// receives a character thru interrupt buffer
int UART_getc(void);

// waits for incoming character in receive buffer directly
int UART_Receive(int wait);

// sends an ASII newline 
void UART_Newline(void);

// prints a NULL terminated string
void UART_Print_String(char *string);

// prints a signed integer
void UART_Print_Int(int num);

// clears the VT100 terminal with ANSI code for clear
void Terminal_Clear_Screen(void);

 // end support for C++ compilers
 #ifdef __cplusplus
 }
 #endif 

// end multiple inclusions 
#endif
