///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Copyright Nurve Networks LLC 2009
// 
// Filename: CHAM_AVR_VGA_STARS_DEMO_01.C
//
// Original Author: Andre' LaMothe
// 
// Last Modified: 10.18.09
//
// Description: VGA Vertical Stars Demo
//  
// Overview: Simply draws a scrolling vertical starfield.
// Chameleon PIC version nearly identical
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
 

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// __AVR_ATmega328P__ , __AVR_ATmega168__
#define __AVR_ATmega328P__ 

// include everything 
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <util/delay.h> 
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <compat/twi.h>
#include <avr/wdt.h>

// include headers required for demo
#include "CHAM_AVR_SYSTEM_V010.h"             // you need this one always
#include "CHAM_AVR_TWI_SPI_DRV_V010.h"        // you need this always
#include "CHAM_AVR_VGA_DRV_V010.h"            // needed for VGA driver

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES AND MACROS /////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// define CPU frequency in Mhz here if not defined in Makefile or other includes, compiler will throw a warning, ignore
#ifndef F_CPU
#define F_CPU 16000000UL // 28636360UL, 14318180UL, 21477270 UL
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MAIN ///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int main(void)
{
int ship_x = 12, ship_y = 0;

// initialize and set SPI rate
SPI_Init( SPI_DEFAULT_RATE );

// give Prop a moment before sending it commands, the boot process is about a second
// if you like you can speed it up, by removing the Prop driver's "LED blink" sequence in the driver
// or speed it up, but commands that are sent before the Prop driver is done booting will be ignored.
// therefore, if you have a lot of set up work to do, then you don't require a delay, but if you jump right
// into commands, then you need a good 1.5 - 2 second delay
_delay_ms( 2500 ); 

// clear screens
VGA_ClearScreen();

// enter infinite loop...
while(1)
     {
     // draw next star

     // set color
     VGA_Color(0);

     // position star at bottom to cause scroll
     VGA_SetXY( rand() % 32, 14 );
     VGA_Term_Print(".                               ");

     // draw our little ship at top
     
     // set color to green (with a little forcefield :)
     VGA_Color(4);
     
     // position at top of screen and draw with ASCII art
     VGA_SetXY( ship_x, ship_y );
     VGA_Term_Print("<-#->");

     // add some logic here to move the ship!

     // slow things down a bit
     _delay_ms(10);

     } // end while    

} // end main
