/* mstuff.c - merlin specific routines */

#include "xlisp.h"
#include "colorimg.h"
#include "mstuff.h"
#include "mdebug.h"
#include "mstubapi.h"
#include "mcom.h"

#ifdef WIN32
#include <direct.h>
#else
#include <unistd.h>
#endif

/* program limits */
#define maxMPECount         32
#define defaultMPECount     4
#define idleTicks           5
#define MAX_LOADPATH_LEN    1024

/* mmp run modes */
#define rmStop  0
#define rmStep  1
#define rmRun   2

/* globals */
char vmlabsPath[MAX_LOADPATH_LEN];

/* local variables */
static int chipRunMode = rmRun;

/* symbols */
xlValue s_FRAME,s_ADDRESS,s_REGISTER,s_OBJECT;
xlValue s_NULL,s_VOID,s_CHAR,s_SHORT,s_INT,s_LONG,s_FLOAT,s_DOUBLE;
xlValue s_STRUCT,s_UNION,s_ENUM,s_MOE,s_UCHAR,s_USHORT,s_UINT,s_ULONG;
xlValue s_POINTER,s_FUNCTION,s_ARRAY;

/* foreign type symbols */
xlValue s_MMP;
xlValue s_MPE;
xlValue s_TYPE;

/* instance variable offsets */
#define connectionHANDLE    xlFIRSTIVAR

/* local variables */
static xlValue s_connection;
static xlValue c_connection;

/* prototypes */
xlValue xversionstring(void);
xlValue xcopyrightstring(void);
void xmmpfeaturecheck(void);
xlValue xconnectMMP(void);
xlValue xmmpstubcontrol(void);
void xmmpcheckstubversion(void);
xlValue xinitMMP(void);
xlValue xresetMMP(void);
xlValue xmmpstart(void);
xlValue xmmpstop(void);
xlValue xmmpMPEcount(void);
xlValue xgetMPE(void);
xlValue xmmpfetchscalar(void);
xlValue xmmpstorescalar(void);
xlValue xmmploadsrecordfile(void);
xlValue xmmpsettimer(void);
xlValue xmmpsendrequest(void);
void xmpeloadobjectfile(void);
void xmpereloadobjectfile(void);
void xmpeloadcofffile(void);
void xmpegetobjectversion(void);
xlValue xmpegetfilereference(void);
void xmpegetbeforemethod(void);
void xmpegetaftermethod(void);
xlValue xmpesetcurrentblock(void);
void xmpefindsymbol(void);
xlValue xmpefindtype(void);
xlValue xmpegettagmembers(void);
xlValue xmpegettypespecifier(void);
xlValue xmpegettypesize(void);
xlValue xmpegetlocalsymbolnames(void);
xlValue xmpegetvalue(void);
xlValue x32bitstofloat(void);
xlValue xfloatto32bits(void);
xlValue x64bitstodouble(void);
void xdoubleto64bits(void);
xlValue xsleep(void);
void xmpefindsymbolsnearestaddress(void);
void xmpefindlinenumber(void);
void xmpefindnextlinenumber(void);
xlValue xmpefindaddressfromlinenumber(void);
void xmpefindfunction(void);
xlValue xmpegetstartaddress(void);
xlValue xmpefetchscalar(void);
xlValue xmpefetchdatascalar(void);
xlValue xmpefetchinstructionscalar(void);
xlValue xmpestorescalar(void);
xlValue xmpestoredatascalar(void);
xlValue xmpestoreinstructionscalar(void);
xlValue xmpefetchstring(void);
xlValue xmpetranslatedataaddress(void);
xlValue xmpetranslateinstructionaddress(void);
xlValue xmmpreadscalarsfromfile(void);
xlValue xmpereadscalarsfromfile(void);
xlValue xmmpwritescalarstofile(void);
xlValue xmpewritescalarstofile(void);
xlValue xmpeinvalidateregistercache(void);
xlValue xmpehookregisterreads(void);
xlValue xmpehookregisterwrites(void);
xlValue xmpeloadregistervalues(void);
void xmpeloadbinaryfile(void);
xlValue xinitmemory(void);
xlValue xmmpsetdisplaymode(void);
xlValue xmmpelapsedticks(void);
void xmmpbusinfo(void);
void xmpeinfo(void);
xlValue xmpestallcount(void);
xlValue xmpepacketcount(void);
xlValue xmperunningcount(void);
xlValue xmpeunitcount(void);
xlValue xmpeinstructioncount(void);
xlValue xmperunningp(void);
xlValue xmpesetstartaddress(void);
xlValue xmpesetsteprange(void);
xlValue xmperun(void);
xlValue xmpestep(void);
xlValue xmpestepover(void);
xlValue xmpestop(void);
xlValue xmperegisteraddress(void);
#ifdef SYM_INST
xlValue xmpestoreinstruction(void);
#endif
xlValue xmmpenablehalthook(void);
xlValue xmpesetbreakpoint(void);
xlValue xmpeclearbreakpoint(void);
xlValue xmpetranslateoverlayaddress(void);
xlValue xmpetranslateloadedoverlayaddress(void);
void xmpepacketinfo(void);
xlValue xmpeinitinstructiondecoder(void);
void xmpedisassembleinstruction(void);
xlValue xdisplaylines(void);
static xlValue xMapOverImageRows(void);
static xlValue xYCrCbToRGB(void);
xlValue xwriteimagetofile(void);
xlValue xwriterawimagetofile(void);
static xlValue xgetcwd(void);
static xlValue xconvertslashes(void);
xlValue xreadmerlinsymbol(void);
#ifdef MSOCKINT
xlValue xmmpinitializesockets(void);
xlValue xmmpfinalizesockets(void);
#endif
xlValue xrunprogram(void);
xlValue xchiprunningp(void);
xlValue xchiprun(void);
xlValue xchipstop(void);
xlValue xchipstep(void);

static xlValue xMComInitialize(void);
static xlValue xMComClose(void);
static xlValue xMComDataAvailableP(void);
static void xMComReadPacket(void);
static xlValue xMComWritePacket(void);

xlSubrDef msubrtab[] = {
{       "VERSION-STRING",                   xversionstring                  },
{       "COPYRIGHT-STRING",                 xcopyrightstring                },
{       "CONNECT-MMP",                      xconnectMMP                     },
{       "MMP-STUB-CONTROL",                 xmmpstubcontrol                 },
{       "INIT-MMP",                         xinitMMP                        },
{       "RESET-MMP",                        xresetMMP                       },
{       "MMP-START",                        xmmpstart                       },
{       "MMP-STOP",                         xmmpstop                        },
{       "MMP-MPE-COUNT",                    xmmpMPEcount                    },
{       "GET-MPE",                          xgetMPE                         },
{       "MMP-FETCH-SCALAR",                 xmmpfetchscalar                 },
{       "MMP-STORE-SCALAR!",                xmmpstorescalar                 },
{       "MMP-READ-SCALARS-FROM-FILE",       xmmpreadscalarsfromfile         },
{       "MMP-WRITE-SCALARS-TO-FILE",        xmmpwritescalarstofile          },
{       "MMP-LOAD-S-RECORD-FILE",           xmmploadsrecordfile             },
{       "MMP-SET-TIMER",                    xmmpsettimer                    },
{       "MPE-GET-FILE-REFERENCE",           xmpegetfilereference            },
{       "MPE-SET-CURRENT-BLOCK!",           xmpesetcurrentblock             },
{       "MPE-FIND-TYPE",                    xmpefindtype                    },
{       "MPE-GET-TAG-MEMBERS",              xmpegettagmembers               },
{       "MPE-GET-TYPE-SPECIFIER",           xmpegettypespecifier            },
{       "MPE-GET-TYPE-SIZE",                xmpegettypesize                 },
{       "MPE-GET-LOCAL-SYMBOL-NAMES",       xmpegetlocalsymbolnames         },
{       "MPE-GET-VALUE",                    xmpegetvalue                    },
{       "32BITS->FLOAT",                    x32bitstofloat                  },
{       "FLOAT->32BITS",                    xfloatto32bits                  },
{       "64BITS->DOUBLE",                   x64bitstodouble                 },
{       "SLEEP",                            xsleep                          },
{       "MPE-GET-START-ADDRESS",            xmpegetstartaddress             },
{       "MPE-FETCH-SCALAR",                 xmpefetchscalar                 },
{       "MPE-FETCH-DATA-SCALAR",            xmpefetchdatascalar             },
{       "MPE-FETCH-INSTRUCTION-SCALAR",     xmpefetchinstructionscalar      },
{       "MPE-FETCH-STRING",                 xmpefetchstring                 },
{       "MPE-STORE-SCALAR!",                xmpestorescalar                 },
{       "MPE-STORE-DATA-SCALAR!",           xmpestoredatascalar             },
{       "MPE-STORE-INSTRUCTION-SCALAR!",    xmpestoreinstructionscalar      },
{       "MPE-TRANSLATE-DATA-ADDRESS",       xmpetranslatedataaddress        },
{       "MPE-TRANSLATE-INSTRUCTION-ADDRESS",xmpetranslateinstructionaddress },
{       "MPE-READ-SCALARS-FROM-FILE",       xmpereadscalarsfromfile         },
{       "MPE-WRITE-SCALARS-TO-FILE",        xmpewritescalarstofile          },
{       "MPE-INVALIDATE-REGISTER-CACHE!",   xmpeinvalidateregistercache     },
{       "MPE-HOOK-REGISTER-READS",          xmpehookregisterreads           },
{       "MPE-HOOK-REGISTER-WRITES",         xmpehookregisterwrites          },
{       "MPE-LOAD-REGISTER-VALUES",         xmpeloadregistervalues          },
{       "INIT-MEMORY!",                     xinitmemory                     },
{       "MMP-SET-DISPLAY-MODE!",            xmmpsetdisplaymode              },
{       "MMP-ELAPSED-TICKS",                xmmpelapsedticks                },
{       "MPE-STALL-COUNT",                  xmpestallcount                  },
{       "MPE-PACKET-COUNT",                 xmpepacketcount                 },
{       "MPE-RUNNING-COUNT",                xmperunningcount                },
{       "MPE-UNIT-COUNT",                   xmpeunitcount                   },
{       "MPE-INSTRUCTION-COUNT",            xmpeinstructioncount            },
{       "MPE-RUNNING?",                     xmperunningp                    },
{       "MPE-SET-START-ADDRESS!",           xmpesetstartaddress             },
{       "MPE-SET-STEP-RANGE!",              xmpesetsteprange                },
{       "MPE-RUN",                          xmperun                         },
{       "MPE-STEP",                         xmpestep                        },
{       "MPE-STEP-OVER",                    xmpestepover                    },
{       "MPE-STOP",                         xmpestop                        },
{       "MPE-REGISTER-ADDRESS",             xmperegisteraddress             },
#ifdef SYM_INST
{       "MPE-STORE-INSTRUCTION!",           xmpestoreinstruction            },
#endif
{       "MPE-SET-BREAKPOINT!",              xmpesetbreakpoint               },
{       "MPE-CLEAR-BREAKPOINT!",            xmpeclearbreakpoint             },
{       "MPE-TRANSLATE-OVERLAY-ADDRESS",    xmpetranslateoverlayaddress     },
{       "MPE-TRANSLATE-LOADED-OVERLAY-ADDRESS",xmpetranslateloadedoverlayaddress },
{       "MPE-FIND-ADDRESS-FROM-LINE-NUMBER",xmpefindaddressfromlinenumber   },
{       "DISPLAY-LINES",                    xdisplaylines                   },
{       "MAP-OVER-IMAGE-ROWS",              xMapOverImageRows               },
{       "YCRCB->RGB",                       xYCrCbToRGB                     },
{       "WRITE-IMAGE-TO-FILE",              xwriteimagetofile               },
{       "WRITE-RAW-IMAGE-TO-FILE",          xwriterawimagetofile            },
{       "GETCWD",                           xgetcwd,                        },
{       "CONVERT-SLASHES",                  xconvertslashes                 },
{       "READ-MERLIN-SYMBOL",               xreadmerlinsymbol               },
{       "MPE-INIT-INSTRUCTION-DECODER",     xmpeinitinstructiondecoder      },
{       "RUN-PROGRAM",                      xrunprogram                     },
{       "CHIP-RUNNING?",                    xchiprunningp                   },
{       "CHIP-RUN",                         xchiprun                        },
{       "CHIP-STOP",                        xchipstop                       },
{       "CHIP-STEP",                        xchipstep                       },
{0,0}};

xlXSubrDef mxsubrtab[] = {
{       "MMP-CHECK-STUB-VERSION",           xmmpcheckstubversion            },
{       "MMP-FEATURE-CHECK",                xmmpfeaturecheck                },
{       "MPE-PACKET-INFO",                  xmpepacketinfo                  },
{       "MPE-DISASSEMBLE-INSTRUCTION",      xmpedisassembleinstruction      },
{       "MPE-LOAD-OBJECT-FILE",             xmpeloadobjectfile              },
{       "MPE-RELOAD-OBJECT-FILE",           xmpereloadobjectfile            },
{       "MPE-LOAD-COFF-FILE",               xmpeloadcofffile                },
{       "MPE-LOAD-BINARY-FILE",             xmpeloadbinaryfile              },
{       "MPE-GET-OBJECT-VERSION",           xmpegetobjectversion            },
{       "MPE-GET-BEFORE-METHOD",            xmpegetbeforemethod             },
{       "MPE-GET-AFTER-METHOD",             xmpegetaftermethod              },
{       "MPE-FIND-SYMBOL",                  xmpefindsymbol                  },
{       "MPE-FIND-SYMBOLS-NEAREST-ADDRESS", xmpefindsymbolsnearestaddress   },
{       "MPE-FIND-LINE-NUMBER",             xmpefindlinenumber              },
{       "MPE-FIND-NEXT-LINE-NUMBER",        xmpefindnextlinenumber          },
{       "MPE-FIND-FUNCTION",                xmpefindfunction                },
{       "MMP-BUS-INFO",                     xmmpbusinfo                     },
{       "MPE-INFO",                         xmpeinfo                        },
{       "DOUBLE->64BITS",                   xdoubleto64bits                 },
{0,0}};

/* numeric constant structure */
typedef struct NumericConstant NumericConstant;
struct NumericConstant {
    char *name;
    int value;
};

/* numeric constants */
static NumericConstant numericConstants[] = {
{   "CRRESTOREDEFAULTS",    crRestoreDefaults   },
{   "CRLOADSETTINGS",       crLoadSettings      },
{   "CRSAVESETTINGS",       crSaveSettings      },
{   "CRGETPOLL",            crGetPoll           },
{   "CRSETPOLL",            crSetPoll           },
{   "CRGETSERVE",           crGetServe          },
{   "CRSETSERVE",           crSetServe          },
{   "CRGETFASTSTEP",        crGetFastStep       },
{   "CRSETFASTSTEP",        crSetFastStep       },
{   "CRGETSTOP",            crGetStop           },
{   "CRSETSTOP",            crSetStop           },
{   "CRGETIPADDR",          crGetIPAddr         },
{   "CRSETIPADDR",          crSetIPAddr         },
{   "CRGETIPMASK",          crGetIPMask         },
{   "CRSETIPMASK",          crSetIPMask         },
{   0,                      0                   }
};

/* local functions */
static xlValue MakeTypeSpecifier(CoffType *type);
static int RegisterReadHook(MPEDebugInfo *mpe,
                            MScalar addr,
                            MScalar value,
                            MScalar *pValue,
                            void *data);
static int RegisterWriteHook(MPEDebugInfo *mpe,
                             MScalar addr,
                             MScalar value,
                             MScalar newValue,
                             MScalar *pValue,
                             void *data);
static void TimerHook(MMPDebugInfo *mmp,void *data);
static void DisplayHook(MMPDebugInfo *mmp,MScalar x,MScalar y,MScalar value);
static void HaltHook(MMPDebugInfo *mnp,int unitNumber,UINT32 reason,MScalar addr,MScalar data);
static void EvalHook(MPEDebugInfo *mpe,char *str);
static void AppendPixelToRow(void *data,int x,int y,int yy,int cr,int cb);
static void MComInitialize(void);
static CommConnection *GetConnection(void);
static void SetConnection(xlValue obj,CommConnection *c);
static CoffType *getTYPE(void);

/* menter - enter merlin specific symbols and functions */
void menter(void)
{
    xlSubrDef *sdp;
    xlXSubrDef *xsdp;
    NumericConstant *c;

    /* enter the built-in functions */
    for (sdp = msubrtab; sdp->name != NULL; ++sdp)
        xlSubr(sdp->name,sdp->subr);
    for (xsdp = mxsubrtab; xsdp->name != NULL; ++xsdp)
        xlXSubr(xsdp->name,xsdp->subr);

    /* enter the storage class symbols */
    s_FRAME    = xlEnter("FRAME");
    s_ADDRESS  = xlEnter("ADDRESS");
    s_REGISTER = xlEnter("REGISTER");
    s_OBJECT   = xlEnter("OBJECT");

    /* enter the data type symbols */
    s_NULL     = xlEnter("NULL");
    s_VOID     = xlEnter("VOID");
    s_CHAR     = xlEnter("CHAR");
    s_SHORT    = xlEnter("SHORT");
    s_INT      = xlEnter("INT");
    s_LONG     = xlEnter("LONG");
    s_FLOAT    = xlEnter("FLOAT");
    s_DOUBLE   = xlEnter("DOUBLE");
    s_STRUCT   = xlEnter("STRUCT");
    s_UNION    = xlEnter("UNION");
    s_ENUM     = xlEnter("ENUM");
    s_MOE      = xlEnter("MEMBER-OF-ENUM");
    s_UCHAR    = xlEnter("UNSIGNED-CHAR");
    s_USHORT   = xlEnter("UNSIGNED-SHORT");
    s_UINT     = xlEnter("UNSIGNED-INT");
    s_ULONG    = xlEnter("UNSIGNED-LONG");
    s_POINTER  = xlEnter("POINTER");
    s_FUNCTION = xlEnter("FUNCTION");
    s_ARRAY    = xlEnter("ARRAY");

    /* enter the foreign pointer types */
    s_MMP  = xlEnter("MMP-HANDLE");
    s_MPE  = xlEnter("MPE-HANDLE");
    s_TYPE = xlEnter("TYPE-HANDLE");

    /* enter the numeric constants */
    for (c = numericConstants; c->name != NULL; ++c) {
        xlValue sym = xlEnter(c->name);
        xlSetValue(sym,xlMakeFixnum(c->value));
    }

    /* initialize the mcom interface */
    MComInitialize();
}

/* mfindsubr - find a merlin specific function */
xlValue (*mfindsubr(char *name))(void)
{
    xlSubrDef *sdp;
    xlXSubrDef *xsdp;

    /* find the built-in function */
    for (sdp = msubrtab; sdp->name != NULL; ++sdp)
        if (strcmp(sdp->name,name) == 0)
            return sdp->subr;
    for (xsdp = mxsubrtab; xsdp->name != NULL; ++xsdp)
        if (strcmp(xsdp->name,name) == 0)
            return (xlValue (*)(void))xsdp->subr;
    return NULL;
}

/* xversionstring - built-in function 'version-string' */
xlValue xversionstring(void)
{
    xlLastArg();
    return xlMakeCString(PUFFIN_VERSION);
}

/* xcopyrightstring - built-in function 'copyright-string' */
xlValue xcopyrightstring(void)
{
    xlLastArg();
    return xlMakeCString(PUFFIN_COPYRIGHT);
}

/* xmmpfeaturecheck - built-in function 'mmp-feature-check' */
void xmmpfeaturecheck(void)
{
    MMPDebugInfo *mmp;
    int value;
    mmp = getMMP();
    xlVal = xlGetArgString();
    xlLastArg();
    if (mmpFeatureCheck(mmp,xlGetString(xlVal),&value)) {
        xlCheck(2);
        xlPush(xlTrue);
        xlPush(value ? xlTrue : xlFalse);
        xlMVReturn(2);
    }
    else
        xlMVReturn(0);
}

/* xconnectMMP - connect to a remote mmp */
xlValue xconnectMMP(void)
{
    MMPDebugInfo *mmp;
    
    /* parse the argument list */
    xlVal = xlGetArgString();
    xlLastArg();

    /* connect to the remote mmp */
    if ((mmp = ConnectMMPDebugger(xlGetString(xlVal))) == NULL)
        return xlNil;

    /* enable the halt and eval hooks */
    mmp->haltHook = HaltHook;
    mmp->evalHook = EvalHook;

    /* return the mmp debugger structure */
    return xlMakeForeignPtr(s_MMP,mmp);
}

/* AttachMMP - attach to an already existing remote mmp connection */
xlValue AttachMMP(MMPDebugInfo *mmp)
{
    return xlMakeForeignPtr(s_MMP,mmp);
}

/* xmmpstubcontrol - built-in function 'MMP-STUB-CONTROL' */
xlValue xmmpstubcontrol(void)
{
    MMPDebugInfo *mmp;
    int fcn,arg=0,value;

    /* parse the argument list */
    mmp = getMMP();
    xlVal = xlGetArgFixnum(); fcn = (int)xlGetFixnum(xlVal);
    if (xlMoreArgsP()) {
        xlVal = xlGetArgFixnum(); arg = (int)xlGetFixnum(xlVal);
    }
    xlLastArg();

    /* arrange to poll the mmp for exceptions */
    return mmpStubControl(mmp,fcn,arg,&value) ? xlMakeFixnum(value) : xlNil;
}

/* xmmpcheckstubversion - check the stub version number */
void xmmpcheckstubversion(void)
{
    MMPDebugInfo *mmp;

    /* parse the argument list */
    mmp = getMMP();
    xlLastArg();

    /* check the stub version number */
    if (mmpCheckStubVersion(mmp)) {
        xlVal = xlTrue;
        xlSVReturn();
    }
    else {
        xlCheck(2);
        xlPush(xlMakeCString(mmp->errorText));
        xlPush(xlFalse);
        xlMVReturn(2);
    }
}

/* xinitMMP - (re)initialize mmp memory */
xlValue xinitMMP(void)
{
    MMPDebugInfo *mmp;

    /* parse the argument list */
    mmp = getMMP();
    xlLastArg();

    /* initialize the mmp debugger */
    InitMMPDebugger(mmp);
    return xlNil;
}

/* xresetMMP - reset mmp memory */
xlValue xresetMMP(void)
{
    MMPDebugInfo *mmp;

    /* parse the argument list */
    mmp = getMMP();
    xlLastArg();

    /* reset the mmp */
    ResetMMP(mmp);
    return xlNil;
}

/* ChipRunningP - check to see if the mmp is running */
int ChipRunningP(void)
{
    return chipRunMode != rmStop;
}

/* ChipRun - start the chip running */
void ChipRun(void)
{
    chipRunMode = rmRun;
}

/* ChipStop - stop the chip from running */
void ChipStop(void)
{
    chipRunMode = rmStop;
}

/* ChipStep - single step the chip */
void ChipStep(void)
{
    chipRunMode = rmStep;
}

/* xmmpstart - built-in function 'MMP-START' */
xlValue xmmpstart(void)
{
    MMPDebugInfo *mmp;

    /* parse the argument list */
    mmp = getMMP();
    xlLastArg();

    /* arrange to poll the mmp for exceptions */
    xlSetIdleHandler((void (*)(void *))mmpPoll,(void *)mmp);
    return xlNil;
}

/* xmmpstop - built-in function 'MMP-STOP' */
xlValue xmmpstop(void)
{
    xlLastArg();
    xlSetIdleHandler(NULL,NULL);
    return xlNil;
}

/* xmmpMPEcount - (re)initialize mmp memory */
xlValue xmmpMPEcount(void)
{
    MMPDebugInfo *mmp;

    /* parse the argument list */
    mmp = getMMP();
    xlLastArg();

    /* return the mpe count */
    return xlMakeFixnum((xlFIXTYPE)mmp->mpeCount);
}

/* xgetMPE - get mpe memory */
xlValue xgetMPE(void)
{
    MPEDebugInfo *mpe;
    MMPDebugInfo *mmp;
    int n;

    /* parse the argument list */
    mmp = getMMP();
    xlVal = xlGetArgFixnum(); n = (int)xlGetFixnum(xlVal);
    xlLastArg();

    /* get the mpe memory */
    mpe = mmpGetMPE(mmp,n);

    /* return the mpe memory structure */
    return mpe ? xlMakeForeignPtr(s_MPE,mpe) : xlNil;
}

/* xmmpfetchscalar - built-in function 'MMP-FETCH-SCALAR' */
xlValue xmmpfetchscalar(void)
{
    MScalar addr,value;
    MMPDebugInfo *mmp;

    /* parse the argument list */
    mmp = getMMP();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* fetch the scalar */
    if (!mmpFetchScalar(mmp,addr,&value))
        return xlNil;

    /* return the value */
    return xlMakeFixnum((xlFIXTYPE)value);
}

/* xmmpstorescalar - built-in function 'MMP-STORE-SCALAR!' */
xlValue xmmpstorescalar(void)
{
    MScalar addr,value;
    MMPDebugInfo *mmp;

    /* parse the argument list */
    mmp = getMMP();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); value = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* store the scalar (in ram) */
    return mmpStoreScalar(mmp,addr,value) ? xlTrue : xlFalse;
}

/* xmmploadsrecordfile - built-in function 'MMP-LOAD-S-RECORD-FILE' */
xlValue xmmploadsrecordfile(void)
{
    MMPDebugInfo *mmp;

    /* parse the argument list */
    mmp = getMMP();
    xlVal = xlGetArgString();
    xlLastArg();

    /* load the s-records */
    return xlMakeFixnum(mmpLoadSRecordFile(mmp,xlGetString(xlVal)));
}

/* xmpeloadobjectfile - built-in function 'MPE-LOAD-OBJECT-FILE' */
void xmpeloadobjectfile(void)
{
    xlValue checkDependenciesP,ignoreBeforeAfterP,useFastLoaderP;
    xlValue loadDebugInfoP,loadCodeP,useBiosLoaderP;
    MPEDebugInfo *mpe;
    int flags = 0;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgString();
    xlGetKeyArg(xlEnterKeyword("CHECK-DEPENDENCIES?"),xlFalse,&checkDependenciesP);
    xlGetKeyArg(xlEnterKeyword("IGNORE-BEFORE-AFTER?"),xlFalse,&ignoreBeforeAfterP);
    xlGetKeyArg(xlEnterKeyword("USE-FAST-LOADER?"),xlTrue,&useFastLoaderP);
    xlGetKeyArg(xlEnterKeyword("LOAD-DEBUGGING-INFORMATION?"),xlTrue,&loadDebugInfoP);
    xlGetKeyArg(xlEnterKeyword("LOAD-CODE?"),xlTrue,&loadCodeP);
    xlGetKeyArg(xlEnterKeyword("USE-BIOS-LOADER?"),xlTrue,&useBiosLoaderP);

    /* build the flags */
    if (checkDependenciesP != xlFalse)
        flags |= lofCheckDependenciesP;
    if (ignoreBeforeAfterP != xlFalse)
        flags |= lofIgnoreBeforeAfterP;
    if (useFastLoaderP != xlFalse)
        flags |= lofUseFastLoaderP;
    if (loadDebugInfoP != xlFalse)
        flags |= lofLoadDebuggingInfoP;
    if (loadCodeP != xlFalse)
        flags |= lofLoadCodeP;
        if (useBiosLoaderP != xlFalse)
                flags |= lofBiosInitOnlyP;

    /* load the object file */
    if (mpeLoadObjectFile(mpe,xlGetString(xlVal),flags)) {
        xlVal = xlTrue;
        xlSVReturn();
    }
    else {
        xlCheck(3);
        xlPush(xlMakeCString(mpe->mmp->errorText));
        xlPush(xlMakeChar(mpe->mmp->errorCode));
        xlPush(xlFalse);
        xlMVReturn(3);
    }
}

/* xmpereloadobjectfile - built-in function 'MPE-RELOAD-OBJECT-FILE' */
void xmpereloadobjectfile(void)
{
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlLastArg();

    /* reload the object file */
    if (mpeReloadObjectFile(mpe)) {
        xlVal = xlTrue;
        xlSVReturn();
    }
    else {
        xlCheck(3);
        xlPush(xlMakeCString(mpe->mmp->errorText));
        xlPush(xlMakeChar(mpe->mmp->errorCode));
        xlPush(xlFalse);
        xlMVReturn(3);
    }
}

/* xmpeloadcofffile - built-in function 'MPE-LOAD-COFF-FILE' */
void xmpeloadcofffile(void)
{
    xlValue ignoreBeforeAfterP,useFastLoaderP,loadDebugInfoP,loadCodeP;
    MPEDebugInfo *mpe;
    int flags = 0;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgString();
    xlGetKeyArg(xlEnterKeyword("IGNORE-BEFORE-AFTER?"),xlFalse,&ignoreBeforeAfterP);
    xlGetKeyArg(xlEnterKeyword("USE-FAST-LOADER?"),xlTrue,&useFastLoaderP);
    xlGetKeyArg(xlEnterKeyword("LOAD-DEBUGGING-INFORMATION?"),xlTrue,&loadDebugInfoP);
    xlGetKeyArg(xlEnterKeyword("LOAD-CODE?"),xlTrue,&loadCodeP);

    /* build the flags */
    if (ignoreBeforeAfterP != xlFalse)
        flags |= lofIgnoreBeforeAfterP;
    if (useFastLoaderP != xlFalse)
        flags |= lofUseFastLoaderP;
    if (loadDebugInfoP != xlFalse)
        flags |= lofLoadDebuggingInfoP;
    if (loadCodeP != xlFalse)
        flags |= lofLoadCodeP;

    /* load the object file */
    if (mpeLoadCoffFile(mpe,xlGetString(xlVal),flags)) {
        xlVal = xlTrue;
        xlSVReturn();
    }
    else {
        xlCheck(3);
        xlPush(xlMakeCString(mpe->mmp->errorText));
        xlPush(xlMakeChar(mpe->mmp->errorCode));
        xlPush(xlFalse);
        xlMVReturn(3);
    }
}

/* xmpeloadbinaryfile - built-in function 'MPE-LOAD-BINARY-FILE' */
void xmpeloadbinaryfile(void)
{
    xlValue useFastLoaderP;
    MPEDebugInfo *mpe;
    int flags = 0;
    MScalar addr;
    long length;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArgString();
    xlGetKeyArg(xlEnterKeyword("USE-FAST-LOADER?"),xlTrue,&useFastLoaderP);

    /* build the flags */
    if (useFastLoaderP != xlFalse)
        flags |= lofUseFastLoaderP;

    /* load the binary file */
    if (!mpeLoadBinaryFile(mpe,addr,xlGetString(xlVal),&length,flags)) {
        xlCPush(xlFalse);
        xlMVReturn(1);
    }
    
    /* return the file length */
    xlCheck(2);
    xlPush(xlMakeFixnum((xlFIXTYPE)length));
    xlPush(xlTrue);
    xlMVReturn(2);
}

/* xmpegetobjectversion - built-in function 'MPE-GET-OBJECT-VERSION' */
void xmpegetobjectversion(void)
{
    char *assemblerVersion;
    long objectFileVersion;
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlLastArg();

    /* get the object version information */
    mpeGetObjectVersion(mpe,&assemblerVersion,&objectFileVersion);

    /* return the version information */
    xlCheck(2);
    xlPush(xlMakeFixnum((xlFIXTYPE)objectFileVersion));
    xlPush(assemblerVersion ? xlMakeCString(assemblerVersion) : xlNil);
    xlMVReturn(2);
}

/* xmpegetfilereference - built-in function 'MPE-GET-FILE-REFERENCE' */
xlValue xmpegetfilereference(void)
{
    char *file;
    MPEDebugInfo *mpe;
    long n;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); n = (long)xlGetFixnum(xlVal);
    xlLastArg();

    /* return the file reference */
    return mpeGetFileReference(mpe,n,&file) ? xlMakeCString(file) : xlNil;
}

/* xmpegetbeforemethod - built-in function 'MPE-GET-BEFORE-METHOD' */
void xmpegetbeforemethod(void)
{
    MPEDebugInfo *mpe;
    MScalar addr;
    char *method;
    long n;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); n = (long)xlGetFixnum(xlVal);
    xlLastArg();

    if (!mpeGetBeforeMethod(mpe,n,&addr,&method)) {
        xlVal = xlNil;
        xlSVReturn();
    }

    /* return the method */
    xlCheck(2);
    xlPush(xlMakeCString(method));
    xlPush(xlMakeFixnum((xlFIXTYPE)addr));
    xlMVReturn(2);
}

/* xmpegetaftermethod - built-in function 'MPE-GET-AFTER-METHOD' */
void xmpegetaftermethod(void)
{
    MPEDebugInfo *mpe;
    MScalar addr;
    char *method;
    long n;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); n = (long)xlGetFixnum(xlVal);
    xlLastArg();

    if (!mpeGetAfterMethod(mpe,n,&addr,&method)) {
        xlVal = xlNil;
        xlSVReturn();
    }

    /* return the method */
    xlCheck(2);
    xlPush(xlMakeCString(method));
    xlPush(xlMakeFixnum((xlFIXTYPE)addr));
    xlMVReturn(2);
}

/* xmpefindsymbol - built-in function 'MPE-FIND-SYMBOL' */
void xmpefindsymbol(void)
{
    MPEDebugInfo *mpe;
    xlValue file,sclass;
    CoffSymbol *s;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgString();
    file = xlMoreArgsP() ? xlGetArg() : xlNil;
    xlLastArg();

    /* find the symbol */
    if (xlNullP(file)) {
        if (!(s = mpeFindSymbol(mpe,xlGetString(xlVal)))) {
            xlVal = xlNil;
            xlSVReturn();
        }
    }

    /* find the symbol in a specific file */
    else if (xlStringP(file)) {
        if (!(s = mpeFindSymbolInFile(mpe,xlGetString(xlVal),xlGetString(file)))) {
            xlVal = xlNil;
            xlSVReturn();
        }
    }
    
    /* bad argument type */
    else
        xlBadType(file);

    /* class */
    switch (s->sclass) {
    case C_AUTO:
    case C_ARG:
        sclass = s_FRAME;
        xlVal = xlMakeFixnum((xlFIXTYPE)s->value);
        break;
    case C_EXT:
    case C_STAT:
    case C_LABEL:
        sclass = s_ADDRESS;
        xlVal = xlMakeFixnum((xlFIXTYPE)s->value);
        break;
    case C_REG:
    case C_REGPARAM:
        sclass = s_REGISTER;
        xlVal = xlMakeFixnum((xlFIXTYPE)s->value);
        break;
    default:
        xlError("unexpected storage class",xlMakeFixnum(s->sclass));
        break;
    }

    /* return the value, overlay, class and type */
    xlCheck(4);
    xlPush(xlMakeForeignPtr(s_TYPE,s->type));
    xlPush(sclass);
    xlPush(s->overlay ? xlMakeFixnum(s->overlay->eAddr) : xlNil);
    xlPush(xlVal);
    xlMVReturn(4);
}

/* xmpesetcurrentblock - built-in function 'MPE-SET-CURRENT-BLOCK!' */
xlValue xmpesetcurrentblock(void)
{
    MPEDebugInfo *mpe;
    MScalar addr;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* set the current block */
    return mpeSetCurrentBlock(mpe,addr,&addr) ? xlMakeFixnum((xlFIXTYPE)addr) : xlFalse;
}

/* xmpefindtype - built-in function 'MPE-FIND-TYPE' */
xlValue xmpefindtype(void)
{
    MPEDebugInfo *mpe;
    CoffSymbol *s;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgString();
    xlLastArg();

    /* find the symbol */
    if (!(s = mpeFindType(mpe,xlGetString(xlVal))))
        return xlNil;

    /* return the type specifier */
    return xlMakeForeignPtr(s_TYPE,s->type);
}

/* xmpegettagmembers - built-in function 'MPE-GET-TAG-MEMBERS' */
xlValue xmpegettagmembers(void)
{
    CoffTagList *tagList;
    MPEDebugInfo *mpe;
    CoffSymbol *s;
    xlValue next;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgString();
    xlLastArg();

    /* find the symbol */
    if (!(s = mpeFindTag(mpe,xlGetString(xlVal))))
        return xlNil;

    /* get the tag list */
    if ((tagList = (CoffTagList *)s->data) == NULL)
        return xlNil;

    /* initialize */
    xlCPush(xlNil);
    next = xlNil;

    /* make a list of members */
    for (s = tagList->tags; s != NULL; s = s->next) {

        /* add the size for bit fields */
        xlVal = xlCons(s->sclass == C_FIELD ? xlMakeFixnum((xlFIXTYPE)s->type->size) : xlFalse,xlNil);

        /* make the type specifier */
        xlVal = xlCons(xlMakeForeignPtr(s_TYPE,s->type),xlVal);

        /* add the offset */
        xlVal = xlCons(xlMakeFixnum((xlFIXTYPE)s->value),xlVal);

        /* add the name */
        xlVal = xlCons(xlMakeCString(s->name),xlVal);

        /* make the list entry */
        xlVal = xlCons(xlVal,xlNil);

        /* add to the list of members */
        if (next == xlNil)
            xlSetTop(xlVal);
        else
            xlSetCdr(next,xlVal);

        /* move ahead */
        next = xlVal;
    }

    /* return the tag list */
    return xlPop();
}

/* xmpegettypespecifier - built-in function 'mpe-get-type-specifier' */
xlValue xmpegettypespecifier(void)
{
    CoffType *type;

    /* parse the argument list */
    type = getTYPE();
    xlLastArg();

    /* make the type specifier */
    return MakeTypeSpecifier(type);
}

/* xmpegettypesize - built-in function 'mpe-get-type-size' */
xlValue xmpegettypesize(void)
{
    CoffType *type;

    /* parse the argument list */
    type = getTYPE();
    xlLastArg();

    /* return the type size */
    return xlMakeFixnum((xlFIXTYPE)type->byteSize);
}

/* MakeTypeSpecifier - make a type specifier from a type */
static xlValue MakeTypeSpecifier(CoffType *type)
{
    /* create type specifier */
    switch (type->type) {
    case T_NULL:
        xlVal = s_NULL;
        break;
    case T_VOID:
        xlVal = s_VOID;
        break;
    case T_CHAR:
        xlVal = s_CHAR;
        break;
    case T_SHORT:
        xlVal = s_SHORT;
        break;
    case T_INT:
        xlVal = s_INT;
        break;
    case T_LONG:
        xlVal = s_LONG;
        break;
    case T_FLOAT:
        xlVal = s_FLOAT;
        break;
    case T_DOUBLE:
        xlVal = s_DOUBLE;
        break;
    case T_STRUCT:
        xlVal = s_STRUCT;
        break;
    case T_UNION:
        xlVal = s_UNION;
        break;
    case T_ENUM:
        xlVal = s_ENUM;
        break;
    case T_MOE:
        xlVal = s_MOE;
        break;
    case T_UCHAR:
        xlVal = s_UCHAR;
        break;
    case T_USHORT:
        xlVal = s_USHORT;
        break;
    case T_UINT:
        xlVal = s_UINT;
        break;
    case T_ULONG:
        xlVal = s_ULONG;
        break;
    case T_POINTER:
        xlVal = s_POINTER;
        break;
    case T_FUNCTION:
        xlVal = s_FUNCTION;
        break;
    case T_ARRAY:
        xlVal = s_ARRAY;
        break;
    default:
        xlError("unexpected type",xlMakeFixnum(type->type));
        break;
    }

    /* make new type specifier */
    switch (type->type) {
    case T_STRUCT:
    case T_UNION:
    case T_ENUM:
        if (type->tag)
            xlVal = xlCons(xlVal,xlCons(xlMakeCString(type->tag->name),xlNil));
        break;
    case T_POINTER:
    case T_FUNCTION:
        xlVal = xlCons(xlVal,xlCons(xlMakeForeignPtr(s_TYPE,type->baseType),xlNil));
        break;
    case T_ARRAY:
        xlCPush(xlCons(xlMakeForeignPtr(s_TYPE,type->baseType),xlNil));
        xlSetTop(xlCons(xlMakeFixnum((xlFIXTYPE)type->size),xlTop()));
        xlVal = xlCons(xlVal,xlPop());
        break;
    }

    /* return the type specifier */
    return xlVal;
}

/* xmpegetlocalsymbolnames - built-in function 'MPE-GET-LOCAL-SYMBOL-NAMES' */
xlValue xmpegetlocalsymbolnames(void)
{
    xlValue newEntry,next;
    MPEDebugInfo *mpe;
    CoffSymbolTable *t;
    CoffSymbol *s;
    int i;
    
    /* parse the argument list */
    mpe = getMPE();
    xlLastArg();

    /* get the local symbol table */
    if ((t = mpeGetLocalSymbolTable(mpe)) == NULL)
        return xlNil;

    /* initialize */
    xlVal = next = xlNil;

    /* loop over each symbol in the table */
    /* the global table has no parent */
    /* file tables have the global table as parent */
    /* local tables have a file table as parent */
    for (; t && t->parent && t->parent->parent; t = t->parent) {
        for (i = 0; i < t->size; ++i) {
            for (s = t->symbols[i]; s != NULL; s = s->next) {
                xlValue pair;

                /* check to see if this symbol is already in the list */
                for (pair = xlVal; pair != xlNil; pair = xlCdr(pair)) {
                    char *name = xlGetString(xlCar(pair));
                    if (strcmp(name,s->name) == 0)
                        goto skip;
                }

                /* make a new list entry */
                newEntry = xlCons(xlMakeCString(s->name),xlNil);

                /* append to the list */
                if (next == xlNil)
                    xlVal = newEntry;
                else
                    xlSetCdr(next,newEntry);

                /* move ahead */
                next = newEntry;

                /* target for skip */
    skip:       ;
            }
        }
    }

    /* return the symbols */
    return xlVal;
}

/* xmpegetvalue - get the value of a c variable */
xlValue xmpegetvalue(void)
{
    MPEDebugInfo *mpe;
    MScalar frame;
    CValue value;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); frame = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArgString();
    xlLastArg();

    /* get the l-value */
    if (!mpeGetLValue(mpe,frame,xlGetString(xlVal),&value))
        return xlNil;

    /* get the r-value */
    if (!mpeGetRValue(mpe,&value))
        return xlNil;

    /* return the value */
    switch (value.type->type) {
    case T_CHAR:
        xlVal = xlMakeFixnum(value.v.charValue);
        break;
    case T_UCHAR:
        xlVal = xlMakeFixnum(value.v.ucharValue);
        break;
    case T_SHORT:
        xlVal = xlMakeFixnum(value.v.shortValue);
        break;
    case T_USHORT:
        xlVal = xlMakeFixnum(value.v.ushortValue);
        break;
    case T_INT:
    case T_MOE:
    case T_LONG:
        xlVal = xlMakeFixnum(value.v.intValue);
        break;
    case T_UINT:
    case T_ULONG:
        xlVal = xlMakeFixnum(value.v.uintValue);
        break;
    case T_FLOAT:
        xlVal = xlMakeFlonum((double)value.v.floatValue);
        break;
    case T_DOUBLE:
        xlVal = xlMakeFlonum(value.v.doubleValue);
        break;
    default:
        xlVal = xlMakeFixnum(value.v.ptrValue);
        break;
    }
    return xlVal;
}

/* x32bitstofloat - built-in function '32BITS->FLOAT' */
xlValue x32bitstofloat(void)
{
    union {
        float fvalue;
        long ivalue;
    } value;

    /* parse the argument list */
    xlVal = xlGetArgFixnum(); value.ivalue = xlGetFixnum(xlVal);
    xlLastArg();

    /* return the float */
    return xlMakeFlonum((xlFLOTYPE)value.fvalue);
}

/* xfloatto32bits - built-in function 'FLOAT->32BITS' */
xlValue xfloatto32bits(void)
{
    union {
        float fvalue;
        long ivalue;
    } value;

    /* parse the argument list */
    xlVal = xlGetArgFlonum(); value.fvalue = (float)xlGetFlonum(xlVal);
    xlLastArg();

    /* return the fixnum */
    return xlMakeFixnum((xlFIXTYPE)value.ivalue);
}

/* x64bitstodouble - built-in function '64BITS->DOUBLE' */
xlValue x64bitstodouble(void)
{
    union {
        double fvalue;
        long ivalue[2];
    } value;

    /* parse the argument list */
    xlVal = xlGetArgFixnum(); value.ivalue[0] = xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); value.ivalue[1] = xlGetFixnum(xlVal);
    xlLastArg();

    /* return the float */
    return xlMakeFlonum((xlFLOTYPE)value.fvalue);
}

/* xdoubleto64bits - built-in function 'DOUBLE->64BITS' */
void xdoubleto64bits(void)
{
    union {
        double fvalue;
        long ivalue[2];
    } value;

    /* parse the argument list */
    xlVal = xlGetArgFlonum(); value.fvalue = (double)xlGetFlonum(xlVal);
    xlLastArg();

    /* return the fixnums */
    xlCheck(2);
    xlPush(xlMakeFixnum((xlFIXTYPE)value.ivalue[1]));
    xlPush(xlMakeFixnum((xlFIXTYPE)value.ivalue[0]));
    xlMVReturn(2);
}

/* xsleep - sleep for a specified number of milliseconds */
xlValue xsleep(void)
{
    xlFIXTYPE msecs;
    xlVal = xlGetArgFixnum(); msecs = xlGetFixnum(xlVal);
    xlLastArg();
#ifdef UNIX
    usleep(msecs * 1000);
#elif WIN32
    Sleep(msecs);
#else
    xlFmtError("sleep not implemented on this platform");
#endif
    return xlNil;
}

/* xmpefindsymbolsnearestaddress - built-in function 'MPE-FIND-SYMBOLS-NEAREST-ADDRESS' */
void xmpefindsymbolsnearestaddress(void)
{
    MPEDebugInfo *mpe;
    MScalar addr,offset;
    char *name;
    int n = 0;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = xlGetFixnum(xlVal);
    xlLastArg();

    /* find the symbol */
    if (!(name = mpeFindSymbolsNearestAddress(mpe,addr,n,&offset))) {
        xlCPush(xlFalse);
        xlMVReturn(1);
    }

    /* build the list of symbol names */
    xlVal = xlCons(xlMakeCString(name),xlNil);
    while ((name = mpeFindSymbolsNearestAddress(mpe,addr,++n,&offset)) != NULL)
        xlVal = xlCons(xlMakeCString(name),xlVal);

    /* return the symbol names and offset */
    xlCheck(2);
    xlPush(xlMakeFixnum((xlFIXTYPE)offset));
    xlPush(xlVal);
    xlMVReturn(2);
}

/* xmpefindlinenumber - built-in function 'MPE-FIND-LINE-NUMBER' */
void xmpefindlinenumber(void)
{
    long file,line,count;
    int offsetP,offset;
    MPEDebugInfo *mpe;
    MScalar addr;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = xlGetFixnum(xlVal);
    offsetP = xlMoreArgsP() && xlGetArg() != xlFalse;
    xlLastArg();

    /* find the line number information */
    if (!mpeFindLineNumber(mpe,addr,&file,&line,&count,offsetP ? &offset : NULL)) {
        xlCPush(xlFalse);
        xlMVReturn(1);
    }

    /* return the line number information and offset */
    xlCheck(4);
    xlPush(offsetP ? xlMakeFixnum((xlFIXTYPE)offset) : xlFalse);
    xlPush(xlMakeFixnum((xlFIXTYPE)count));
    xlPush(xlMakeFixnum((xlFIXTYPE)line));
    xlPush(xlMakeFixnum((xlFIXTYPE)file));
    xlMVReturn(4);
}

/* xmpefindnextlinenumber - built-in function 'MPE-FIND-NEXT-LINE-NUMBER' */
void xmpefindnextlinenumber(void)
{
    long file,line,count;
    MPEDebugInfo *mpe;
    MScalar addr;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = xlGetFixnum(xlVal);
    xlLastArg();

    /* find the line number information */
    if (!mpeFindNextLineNumber(mpe,addr,&file,&line,&count)) {
        xlCPush(xlFalse);
        xlMVReturn(1);
    }

    /* return the line number information */
    xlCheck(3);
    xlPush(xlMakeFixnum((xlFIXTYPE)count));
    xlPush(xlMakeFixnum((xlFIXTYPE)line));
    xlPush(xlMakeFixnum((xlFIXTYPE)file));
    xlMVReturn(3);
}

/* xmpefindaddressfromlinenumber - built-in function 'MPE-FIND-ADDRESS-FROM-LINE-NUMBER' */
xlValue xmpefindaddressfromlinenumber(void)
{
    MPEDebugInfo *mpe;
    long file,line;
    MScalar addr;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); file = xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); line = xlGetFixnum(xlVal);
    xlLastArg();

    /* find the line number information */
    if (!mpeFindAddressFromLineNumber(mpe,file,line,&addr))
        return xlNil;

    /* return the address */
    return xlMakeFixnum((xlFIXTYPE)addr);
}

/* xmpefindfunction - built-in function 'MPE-FIND-FUNCTION' */
void xmpefindfunction(void)
{
    MScalar addr,baseAddr;
    MPEDebugInfo *mpe;
    char *name;
    long file;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = xlGetFixnum(xlVal);
    xlLastArg();

    /* find the function */
    if (!mpeFindFunction(mpe,addr,&file,&name,&baseAddr)) {
        xlCPush(xlFalse);
        xlMVReturn(1);
    }

    /* return the file and name */
    xlCheck(3);
    xlPush(xlMakeFixnum((xlFIXTYPE)baseAddr));
    xlPush(xlMakeCString(name));
    xlPush(xlMakeFixnum((xlFIXTYPE)file));
    xlMVReturn(3);
}

/* xmpegetstartaddress - built-in function 'MPE-GET-START-ADDRESS' */
xlValue xmpegetstartaddress(void)
{
    MScalar startAddress;
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlLastArg();

    /* return the start address */
    return mpeGetStartAddress(mpe,&startAddress) ? xlMakeFixnum((xlFIXTYPE)startAddress)
                                                 : xlNil;
}

/* xmpefetchscalar - built-in function 'MPE-FETCH-SCALAR' */
xlValue xmpefetchscalar(void)
{
    MScalar addr,value;
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* fetch the scalar */
    if (!mpeFetchScalar(mpe,addr,&value))
        return xlNil;

    /* return the value */
    return xlMakeFixnum((xlFIXTYPE)value);
}

/* xmpefetchdatascalar - built-in function 'MPE-FETCH-DATA-SCALAR' */
xlValue xmpefetchdatascalar(void)
{
    MScalar addr,value;
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* fetch the scalar */
    if (!mpeFetchDataScalar(mpe,addr,&value))
        return xlNil;

    /* return the value */
    return xlMakeFixnum((xlFIXTYPE)value);
}

/* xmpefetchinstructionscalar - built-in function 'MPE-FETCH-INSTRUCTION-SCALAR' */
xlValue xmpefetchinstructionscalar(void)
{
    MScalar addr,value;
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* fetch the scalar */
    if (!mpeFetchInstructionScalar(mpe,addr,&value))
        return xlNil;

    /* return the value */
    return xlMakeFixnum((xlFIXTYPE)value);
}

/* xmpestorescalar - built-in function 'MPE-STORE-SCALAR!' */
xlValue xmpestorescalar(void)
{
    MScalar addr,value;
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); value = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* store the scalar (in ram) */
    return mpeStoreScalar(mpe,addr,value) ? xlTrue : xlFalse;
}

/* xmpestoredatascalar - built-in function 'MPE-STORE-DATA-SCALAR!' */
xlValue xmpestoredatascalar(void)
{
    MScalar addr,value;
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); value = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* store the scalar (in ram) */
    return mpeStoreDataScalar(mpe,addr,value) ? xlTrue : xlFalse;
}

/* xmpestoreinstructionscalar - built-in function 'MPE-STORE-INSTRUCTION-SCALAR!' */
xlValue xmpestoreinstructionscalar(void)
{
    MScalar addr,value;
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); value = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* store the scalar (in ram) */
    return mpeStoreInstructionScalar(mpe,addr,value) ? xlTrue : xlFalse;
}

/* FetchByte cache */
static MScalar fetchByteAddr;
static MScalar fetchByteValue;
static int fetchByteValidP;
static int fetchByteUseCacheP;

/* InitFetchByteCache - initialize the fetch byte cache */
static void InitFetchByteCache(int useCacheP)
{
    fetchByteUseCacheP = useCacheP;
    fetchByteValidP = FALSE;
}

/* FetchByte - fetch a byte using a one scalar cache */
static int FetchByte(MPEDebugInfo *mpe,MScalar addr,int *pByte)
{
    MScalar scalarAddr = addr & scalarAddressMask;

    /* check to see if the scalar is already in the cache */
    if (!fetchByteValidP || scalarAddr != fetchByteAddr) {
        
        /* fetch the scalar containing the requested byte */
        if (fetchByteUseCacheP) {
            if (!mpeFetchDataScalar(mpe,scalarAddr,&fetchByteValue))
                return FALSE;
        }
        else {
            if (!mpeFetchScalar(mpe,scalarAddr,&fetchByteValue))
                return FALSE;
        }

        /* mark the cache as valid */
        fetchByteAddr = scalarAddr;
        fetchByteValidP = TRUE;
    }

    /* return the requested byte */
    *pByte = (fetchByteValue >> (24 - (addr & 3) * 8)) & 0xff;
    return TRUE;
}

/* mpefetchstring - fetch a string */
xlValue xmpefetchstring(void)
{
    xlValue countArg,useCachePArg;
    MScalar addr,count,length,src;
    int countP,useCacheP,ch;
    MPEDebugInfo *mpe;
    char *dst;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlGetKeyArg(xlEnterKeyword("COUNT"),xlNil,&countArg);
    xlGetKeyArg(xlEnterKeyword("USE-CACHE?"),xlTrue,&useCachePArg);

    /* decode the keyword arguments */
    if (countArg == xlNil)
        countP = FALSE;
    else {
        count = (MScalar)xlGetFixnum(countArg);
        countP = TRUE;
    }
    useCacheP = useCachePArg != xlFalse;

    /* first determine the length of the string */
    InitFetchByteCache(useCacheP);
    for (src = addr, length = 0; !countP || --count >= 0; ++src)
        if (FetchByte(mpe,src,&ch) && ch != '\0')
            ++length;
        else
            break;

    /* allocate the new string */
    xlVal = xlNewString(length);

    /* fetch the string */
    InitFetchByteCache(useCacheP);
    for (src = addr, dst = xlGetString(xlVal); --length >= 0; ++src)
        if (FetchByte(mpe,src,&ch))
            *dst++ = ch;

    /* return the new string */
    return xlVal;
}

/* xmpetranslatedataaddress - built-in function 'MPE-TRANSLATE-DATA-ADDRESS */
xlValue xmpetranslatedataaddress(void)
{
    MPEDebugInfo *mpe;
    MScalar addr;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* return the translated address */
    return mpeTranslateDataAddress(mpe,addr,&addr) ? xlMakeFixnum((xlFIXTYPE)addr) : xlNil;
}

/* xmpetranslateinstructionaddress - built-in function 'MPE-TRANSLATE-INSTRUCTION-ADDRESS */
xlValue xmpetranslateinstructionaddress(void)
{
    MPEDebugInfo *mpe;
    MScalar addr;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* return the translated address */
    return mpeTranslateInstructionAddress(mpe,addr,&addr) ? xlMakeFixnum((xlFIXTYPE)addr) : xlNil;
}

#define BlockSize   1024

/* NuonToHost - swap bytes from nuon byte order to host byte order */
static void NuonToHost(MScalar *buf,int cnt)
{
    while (--cnt >= 0) {
        *buf = ntohl(*buf);
        ++buf;
    }
}

/* xmmpreadscalarsfromfile - built-in function 'MMP-READ-SCALARS-FROM-FILE' */
xlValue xmmpreadscalarsfromfile(void)
{
    MScalar addr,buf[BlockSize];
    MMPDebugInfo *mmp;
    int count,cnt;
    FILE *fp;

    /* parse the argument list */
    mmp = getMMP();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); count = (int)xlGetFixnum(xlVal);
    fp = xlGetFile(xlMoreArgsP() ? xlGetInputPort() : xlCurInput());
    xlLastArg();

    /* fetch the scalars */
    while (count > 0) {

        /* determine the size of this block */
        if ((cnt = count) > BlockSize)
            cnt = BlockSize;

        /* read this block from the file */
        if (xlosReadFile(buf,sizeof(MScalar),cnt,fp) != (size_t)cnt)
            return xlFalse;

        /* swap the bytes to host byte order */
        NuonToHost(buf,cnt);

        /* store this block */
        if (!mmpStoreScalars(mmp,addr,cnt,buf))
            return xlFalse;

        /* update */
        addr += cnt * sizeof(MScalar);
        count -= cnt;
    }

    /* return successfully */
    return xlTrue;
}

/* xmpereadscalarsfromfile - built-in function 'MPE-READ-SCALARS-FROM-FILE' */
xlValue xmpereadscalarsfromfile(void)
{
    MScalar addr,buf[BlockSize];
    MPEDebugInfo *mpe;
    int count,cnt;
    FILE *fp;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); count = (int)xlGetFixnum(xlVal);
    fp = xlGetFile(xlMoreArgsP() ? xlGetInputPort() : xlCurInput());
    xlLastArg();

    /* fetch the scalars */
    while (count > 0) {

        /* determine the size of this block */
        if ((cnt = count) > BlockSize)
            cnt = BlockSize;

        /* read this block from the file */
        if (xlosReadFile(buf,sizeof(MScalar),cnt,fp) != (size_t)cnt)
            return xlFalse;

        /* swap the bytes to host byte order */
        NuonToHost(buf,cnt);

        /* store this block */
        if (!mpeStoreScalars(mpe,addr,cnt,buf))
            return xlFalse;

        /* update */
        addr += cnt * sizeof(MScalar);
        count -= cnt;
    }

    /* return successfully */
    return xlTrue;
}

/* HostToNuon - swap bytes from host byte order to nuon byte order */
static void HostToNuon(MScalar *buf,int cnt)
{
    while (--cnt >= 0) {
        *buf = htonl(*buf);
        ++buf;
    }
}

/* xmmpwritescalarstofile - built-in function 'MMP-WRITE-SCALARS-TO-FILE' */
xlValue xmmpwritescalarstofile(void)
{
    MScalar addr,buf[BlockSize];
    MMPDebugInfo *mmp;
    int count,cnt;
    FILE *fp;

    /* parse the argument list */
    mmp = getMMP();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); count = (int)xlGetFixnum(xlVal);
    fp = xlGetFile(xlMoreArgsP() ? xlGetOutputPort() : xlCurOutput());
    xlLastArg();

    /* fetch the scalars */
    while (count > 0) {

        /* determine the size of this block */
        if ((cnt = count) > BlockSize)
            cnt = BlockSize;

        /* fetch this block */
        if (!mmpFetchScalars(mmp,addr,cnt,buf))
            return xlFalse;

        /* swap the bytes to nuon byte order */
        HostToNuon(buf,cnt);

        /* write this block to the file */
        if (xlosWriteFile(buf,sizeof(MScalar),cnt,fp) != (size_t)cnt)
            return xlFalse;

        /* update */
        addr += cnt * sizeof(MScalar);
        count -= cnt;
    }

    /* return successfully */
    return xlTrue;
}

/* xmpewritescalarstofile - built-in function 'MPE-WRITE-SCALARS-TO-FILE' */
xlValue xmpewritescalarstofile(void)
{
    MScalar addr,buf[BlockSize];
    MPEDebugInfo *mpe;
    int count,cnt;
    FILE *fp;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); count = (int)xlGetFixnum(xlVal);
    fp = xlGetFile(xlMoreArgsP() ? xlGetOutputPort() : xlCurOutput());
    xlLastArg();

    /* fetch the scalars */
    while (count > 0) {

        /* determine the size of this block */
        if ((cnt = count) > BlockSize)
            cnt = BlockSize;

        /* fetch this block */
        if (!mpeFetchScalars(mpe,addr,cnt,buf))
            return xlFalse;

        /* swap the bytes to nuon byte order */
        HostToNuon(buf,cnt);

        /* write this block to the file */
        if (xlosWriteFile(buf,sizeof(MScalar),cnt,fp) != (size_t)cnt)
            return xlFalse;

        /* update */
        addr += cnt * sizeof(MScalar);
        count -= cnt;
    }

    /* return successfully */
    return xlTrue;
}

/* xmpeinvalidateregistercache - built-in function 'MPE-INVALIDATE-REGISTER-CACHE!' */
xlValue xmpeinvalidateregistercache(void)
{
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlLastArg();

    /* invalidate the register cache */
    mpeInvalidateRegisterCache(mpe);
    return xlNil;
}

/* callback data structure */
typedef struct {
    xlValue handler;
} CallbackData;

/* RegisterReadHook - register read hook function */
static int RegisterReadHook(MPEDebugInfo *mpe,
                            MScalar addr,
                            MScalar value,
                            MScalar *pValue,
                            void *data)
{
    CallbackData *cb = (CallbackData *)data;
    xlValue vMPE,vAddr,vValue,values[2];
    int vcnt;
    xlCheck(2);
    xlPush(xlMakeForeignPtr(s_MPE,mpe));
    xlPush(xlMakeFixnum((xlFIXTYPE)addr));
    vValue = xlMakeFixnum((xlFIXTYPE)value);
    vAddr = xlPop();
    vMPE = xlPop();
    vcnt = xlCallFunction(values,2,cb->handler,3,vMPE,vAddr,vValue);
    if (xlFixnumP(values[0]))
        *pValue = (MScalar)xlGetFixnum(xlVal);
    return vcnt > 1 && values[1] == xlTrue;
}

/* xmpehookregisterreads - built-in function 'MPE-HOOK-REGISTER-READS' */
xlValue xmpehookregisterreads(void)
{
    CallbackData *cb;
    MPEDebugInfo *mpe;
    MScalar addr;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArg();
    xlLastArg();

    /* allocate callback structure */
    if ((cb = (CallbackData *)malloc(sizeof(CallbackData))) == NULL)
        return xlFalse;

    /* setup the callback */
    if (!mpeHookRegisterReads(mpe,addr,RegisterReadHook,cb)) {
        free(cb);
        return xlFalse;
    }

    /* initialize the callback structure */
    xlProtect(&cb->handler);
    cb->handler = xlVal;

    /* return successfully */
    return xlTrue;
}

/* RegisterWriteHook - register write hook function */
static int RegisterWriteHook(MPEDebugInfo *mpe,
                             MScalar addr,
                             MScalar value,
                             MScalar newValue,
                             MScalar *pValue,
                             void *data)
{
    CallbackData *cb = (CallbackData *)data;
    xlValue vMPE,vAddr,vValue,vNewValue,values[2];
    int vcnt;
    xlCheck(3);
    xlPush(xlMakeForeignPtr(s_MPE,mpe));
    xlPush(xlMakeFixnum((xlFIXTYPE)addr));
    xlPush(xlMakeFixnum((xlFIXTYPE)value));
    vNewValue = xlMakeFixnum((xlFIXTYPE)newValue);
    vValue = xlPop();
    vAddr = xlPop();
    vMPE = xlPop();
    vcnt = xlCallFunction(values,2,cb->handler,4,vMPE,vAddr,vValue,vNewValue);
    if (xlFixnumP(values[0]))
        *pValue = (MScalar)xlGetFixnum(xlVal);
    return vcnt > 1 && values[1] == xlTrue;
}

/* xmpehookregisterwrites - built-in function 'MPE-HOOK-REGISTER-WRITES' */
xlValue xmpehookregisterwrites(void)
{
    CallbackData *cb;
    MPEDebugInfo *mpe;
    MScalar addr;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArg();
    xlLastArg();

    /* allocate callback structure */
    if ((cb = (CallbackData *)malloc(sizeof(CallbackData))) == NULL)
        return xlFalse;

    /* setup the callback */
    if (!mpeHookRegisterWrites(mpe,addr,RegisterWriteHook,cb)) {
        free(cb);
        return xlFalse;
    }

    /* initialize the callback structure */
    xlProtect(&cb->handler);
    cb->handler = xlVal;

    /* return successfully */
    return xlTrue;
}

/* xmpeloadregistervalues - built-in function 'MPE-LOAD-REGISTER-VALUES' */
xlValue xmpeloadregistervalues(void)
{
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlLastArg();

    /* load register values into the register cache */
    mpeLoadRegisterValues(mpe,TRUE);

    /* return successfully */
    return xlTrue;
}

/* TimerHook - timer hook function */
static void TimerHook(MMPDebugInfo *mmp,void *data)
{
    CallbackData *cb = (CallbackData *)data;
    xlCallFunction(NULL,0,cb->handler,1,xlMakeForeignPtr(s_MMP,mmp));
    xlUnprotect(&cb->handler);
    free(cb);
}

/* xmmpsettimer - built-in function 'MMP-SET-TIMER' */
xlValue xmmpsettimer(void)
{
    MMPDebugInfo *mmp;
    MScalar tickCount;
    CallbackData *cb;

    /* parse the argument list */
    mmp = getMMP();
    xlVal = xlGetArgFixnum(); tickCount = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArg();
    xlLastArg();

    /* allocate callback structure */
    if ((cb = (CallbackData *)malloc(sizeof(CallbackData))) == NULL)
        return xlFalse;

    /* setup the callback */
    if (!mmpSetTimer(mmp,tickCount,TimerHook,cb)) {
        free(cb);
        return xlFalse;
    }

    /* initialize the callback structure */
    xlProtect(&cb->handler);
    cb->handler = xlVal;

    /* return successfully */
    return xlTrue;
}

/* DisplayHook - routine to update a pixel */
static void DisplayHook(MMPDebugInfo *mmp,MScalar x,MScalar y,MScalar value)
{
    Color color;
    MapYCrCbToRGB((value >> 24) & 0xff,
                  (value >> 16) & 0xff,
                  (value >>  8) & 0xff,
                  &color);
    xlVal = xlMakeFixnum((xlFIXTYPE)( color.r
                                    |(color.g <<  8)
                                    |(color.b << 16)));
    xlCallFunctionByName(NULL,0,"GG-UPDATE-PIXEL",3,xlMakeFixnum(x),xlMakeFixnum(y),xlVal);
}

/* xmmpsetdisplaymode - set the display mode of an MMP */
xlValue xmmpsetdisplaymode(void)
{
    MScalar base,width,height,mode;
    MMPDebugInfo *mmp;

    /* parse the argument list */
    mmp = getMMP();
    xlVal = xlGetArgFixnum(); base = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); width = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); height = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); mode = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* enable the display hook */
    mmp->displayHook = DisplayHook;

    /* set the display mode */
    return mmpSetDisplayMode(mmp,base,width,height,mode) ? xlTrue : xlFalse;
}

/* xinitmemory - initialize memory */
xlValue xinitmemory(void)
{
    MScalar addr,count,values[4];
    int valueCount = 0;
    MMPDebugInfo *mmp;

    /* parse the argument list */
    mmp = getMMP();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); count = (MScalar)xlGetFixnum(xlVal);
    while (xlMoreArgsP() && valueCount < 4) {
        xlVal = xlGetArgFixnum(); values[valueCount] = (MScalar)xlGetFixnum(xlVal);
        ++valueCount;
    }
    xlLastArg();

    /* initialize to zero if no values were specified */
    if (valueCount == 0) {
        valueCount = 1;
        values[0] = 0;
    }

    /* initialize memory */
    return mmpInitMemory(mmp,addr,count,valueCount,values) ? xlTrue : xlFalse;
}

/* xchiprunningp - built-in function 'CHIP-RUNNING?' */
xlValue xchiprunningp(void)
{
    xlLastArg();
    return ChipRunningP() ? xlTrue : xlFalse;
}

/* xchiprun - built-in function 'CHIP-RUN' */
xlValue xchiprun(void)
{
    xlLastArg();
    ChipRun();
    return xlNil;
}

/* xchipstop - built-in function 'CHIP-STOP' */
xlValue xchipstop(void)
{
    xlLastArg();
    ChipStop();
    return xlNil;
}

/* xchipstep - built-in function 'CHIP-STEP' */
xlValue xchipstep(void)
{
    xlLastArg();
    ChipStep();
    return xlNil;
}

/* xmmpelapsedticks - built-in function 'MMP-ELAPSED-TICKS' */
xlValue xmmpelapsedticks(void)
{
    MMPDebugInfo *mmp;
    MMPInfo info;

    /* parse the argument list */
    mmp = getMMP();
    xlLastArg();

    /* get the number of elapsed ticks */
    if (!mmpGetInfo(mmp,&info))
        return xlNil;

    /* return the elapsed ticks */
    return xlMakeFixnum((xlFIXTYPE)info.elapsedTicks);
}

/* xmmpbusinfo - built-in function 'MMP-BUS-INFO' */
void xmmpbusinfo(void)
{
    MMPDebugInfo *mmp;
    MMPInfo info;

    /* parse the argument list */
    mmp = getMMP();
    xlLastArg();

    /* get the bus information */
    if (!mmpGetInfo(mmp,&info)) {
        xlVal = xlNil;
        xlSVReturn();
    }

    /* return bus information */
    xlCheck(6);
#ifdef BETA
    xlPush(xlMakeFixnum(info.otherDataCycles));
    xlPush(xlMakeFixnum(info.otherDecodeCycles));
    xlPush(xlMakeFixnum(info.otherCommandCycles));
    xlPush(xlMakeFixnum(info.otherIdleCycles));
#endif
    xlPush(xlMakeFixnum(info.dataCycles));
    xlPush(xlMakeFixnum(info.decodeCycles));
    xlPush(xlMakeFixnum(info.commandCycles));
    xlPush(xlMakeFixnum(info.burstVideoCycles));
    xlPush(xlMakeFixnum(info.videoCycles));
    xlPush(xlMakeFixnum(info.idleCycles));
    xlMVReturn(6);
}

/* xmpestallcount - built-in function 'MPE-STALL-COUNT' */
xlValue xmpestallcount(void)
{
    MPEDebugInfo *mpe;
    MPEInfo info;

    /* parse the argument list */
    mpe = getMPE();
    xlLastArg();

    /* get the mpe info */
    if (!mpeGetInfo(mpe,&info))
        return xlNil;

    /* return the stall count */
    return xlMakeFixnum((xlFIXTYPE)info.stallCount);
}

/* xmpepacketcount - built-in function 'MPE-PACKET-COUNT' */
xlValue xmpepacketcount(void)
{
    MPEDebugInfo *mpe;
    MPEInfo info;

    /* parse the argument list */
    mpe = getMPE();
    xlLastArg();

    /* get the mpe info */
    if (!mpeGetInfo(mpe,&info))
        return xlNil;

    /* return the packet count */
    return xlMakeFixnum((xlFIXTYPE)info.packetCount);
}

/* xmperunningcount - built-in function 'MPE-RUNNING-COUNT' */
xlValue xmperunningcount(void)
{
    MPEDebugInfo *mpe;
    MPEInfo info;

    /* parse the argument list */
    mpe = getMPE();
    xlLastArg();

    /* get the mpe info */
    if (!mpeGetInfo(mpe,&info))
        return xlNil;

    /* return the packet count */
    return xlMakeFixnum((xlFIXTYPE)info.runningCount);
}

/* xmpeinfo - built-in function 'MPE-INFO' */
void xmpeinfo(void)
{
    MPEDebugInfo *mpe;
    MPEInfo info;

    /* parse the argument list */
    mpe = getMPE();
    xlLastArg();

    /* get the mpe info */
    if (!mpeGetInfo(mpe,&info)) {
        xlVal = xlNil;
        xlSVReturn();
    }

    /* return bus information */
    xlCheck(3);
    xlPush(xlMakeFixnum(info.runningCount));
    xlPush(xlMakeFixnum(info.packetCount));
    xlPush(xlMakeFixnum(info.stallCount));
    xlMVReturn(3);
}

/* xmpeinstructioncount - built-in function 'MPE-INSTRUCTION-COUNT' */
xlValue xmpeinstructioncount(void)
{
    MPEDebugInfo *mpe;
    UINT32 count;
    int i;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); i = (int)xlGetFixnum(xlVal);
    xlLastArg();

    /* return the instruction count */
    return mpeInstructionCount(mpe,i,&count) ? xlMakeFixnum((xlFIXTYPE)count) : xlNil;
}

/* xmpeunitcount - built-in function 'MPE-UNIT-COUNT' */
xlValue xmpeunitcount(void)
{
    MPEDebugInfo *mpe;
    xlFIXTYPE size,i;
    UINT32 count;
    xlValue val;

    /* parse the argument list */
    mpe = getMPE();
    xlLastArg();

    /* get the function unit counts */
    for (size = 0; mpeUnitCount(mpe,size,&count); ++size)
        xlVal = xlCons(xlMakeFixnum(count),xlVal);

    /* make the result vector */
    val = xlNewVector(size);

    /* fill in the function unit counts */
    for (i = size; --i >= 0; xlVal = xlCdr(xlVal))
        xlSetElement(val,i,xlCar(xlVal));

    /* return the vector of counts */
    return val;
}

/* xmperunningp - built-in function 'MPE-RUNNING?' */
xlValue xmperunningp(void)
{
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlLastArg();

    /* return the running state */
    return mpeRunningP(mpe) ? xlTrue : xlFalse;
}

/* xmpesetstartaddress - built-in function 'MPE-SET-START-ADDRESS!' */
xlValue xmpesetstartaddress(void)
{
    MPEDebugInfo *mpe;
    MScalar addr;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = xlGetFixnum(xlVal);
    xlLastArg();

    /* set the start address */
    mpeSetStartAddress(mpe,addr);
    return xlNil;
}

/* xmpesetsteprange - built-in function 'MPE-SET-STEP-RANGE!' */
xlValue xmpesetsteprange(void)
{
    MPEDebugInfo *mpe;
    MScalar start,end;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); start = xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); end = xlGetFixnum(xlVal);
    xlLastArg();

    /* set the start address */
    mpeSetStepRange(mpe,start,end);
    return xlNil;
}

/* xmperun - built-in function 'MPE-RUN' */
xlValue xmperun(void)
{
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlLastArg();

    /* start the mpe running */
    mpeRun(mpe);
    return xlNil;
}

/* xmpestep - built-in function 'MPE-STEP' */
xlValue xmpestep(void)
{
    MPEDebugInfo *mpe;
    int stepOverIntP;

    /* parse the argument list */
    mpe = getMPE();
    stepOverIntP = xlMoreArgsP() && xlGetArg() != xlFalse;
    xlLastArg();

    /* step the mpe */
    mpeStep(mpe,stepOverIntP);
    return xlNil;
}

/* xmpestepover - built-in function 'MPE-STEP-OVER' */
xlValue xmpestepover(void)
{
    MPEDebugInfo *mpe;
    int stepOverIntP;

    /* parse the argument list */
    mpe = getMPE();
    stepOverIntP = xlMoreArgsP() && xlGetArg() != xlFalse;
    xlLastArg();

    /* step the mpe */
    mpeStepOver(mpe,stepOverIntP);
    return xlNil;
}

/* xmpestop - built-in function 'MPE-STOP' */
xlValue xmpestop(void)
{
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlLastArg();

    /* stop the mpe from running */
    mpeStop(mpe);
    return xlNil;
}

/* xmperegisteraddress - built-in function 'MPE-REGISTER-ADDRESS' */
xlValue xmperegisteraddress(void)
{
    MPEDebugInfo *mpe;
    MScalar addr;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgString();
    xlLastArg();

    /* get the register value */
    if (!mpeRegisterAddress(mpe,xlGetString(xlVal),&addr))
        return xlNil;

    /* return the address */
    return xlMakeFixnum((xlFIXTYPE)addr);
}

#ifdef SYM_INST
/* xmpestoreinstruction - built-in function 'MPE-STORE-INSTRUCTION!' */
xlValue xmpestoreinstruction(void)
{
    MScalar addr,flags;
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); flags = xlGetFixnum(xlVal);
    xlVal = xlGetArgString();
    xlLastArg();

    /* store the instruction */
    return xlMakeFixnum((xlFIXTYPE)mpeStoreInstruction(mpe,addr,flags,xlGetString(xlVal)));
}
#endif

/* HaltHook - internal halt hook handler */
static void HaltHook(MMPDebugInfo *mmp,int unitNumber,UINT32 reason,MScalar addr,MScalar data)
{
    xlValue vMMP,vUnitNumber,vReason,vAddr,vData;
    xlCheck(4);
    xlPush(xlMakeForeignPtr(s_MMP,mmp));
    xlPush(xlMakeFixnum((xlFIXTYPE)unitNumber));
    xlPush(xlMakeFixnum((xlFIXTYPE)reason));
    xlPush(xlMakeFixnum((xlFIXTYPE)addr));
    xlPush(xlMakeFixnum((xlFIXTYPE)data));
    vData = xlPop();
    vAddr = xlPop();
    vReason = xlPop();
    vUnitNumber = xlPop();
    vMMP = xlPop();
    xlCallFunctionByName(NULL,0,"HALT-HOOK",5,vMMP,vUnitNumber,vReason,vAddr,vData);
    xlosFlushOutput();
}

/* EvalHook - internal eval hook handler */
static void EvalHook(MPEDebugInfo *mpe,char *str)
{
    xlValue vMPE,vStr;
    xlCPush(xlMakeForeignPtr(s_MPE,mpe));
    vStr = xlMakeCString(str);
    vMPE = xlPop();
    xlCallFunctionByName(NULL,0,"EVAL-HOOK",2,vMPE,vStr);
}

/* xmpesetbreakpoint - built-in function 'MPE-SET-BREAKPOINT!' */
xlValue xmpesetbreakpoint(void)
{
    MPEDebugInfo *mpe;
    MScalar addr;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* set a breakpoint */
    return mpeSetBreakpoint(mpe,addr) ? xlTrue : xlFalse;
}

/* xmpeclearbreakpoint - built-in function 'MPE-CLEAR-BREAKPOINT!' */
xlValue xmpeclearbreakpoint(void)
{
    MPEDebugInfo *mpe;
    MScalar addr;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* clear a breakpoint */
    return mpeClearBreakpoint(mpe,addr) ? xlTrue : xlFalse;
}

/* xmpetranslateoverlayaddress - built-in function 'MPE-TRANSLATE-OVERLAY-ADDRESS' */
xlValue xmpetranslateoverlayaddress(void)
{
    MScalar addr,overlay;
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); overlay = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* translate the address */
    if (!mpeTranslateOverlayAddress(mpe,overlay,addr,&addr))
        return xlNil;

    /* return the translated address */
    return xlMakeFixnum((xlFIXTYPE)addr);
}

/* xmpetranslateloadedoverlayaddress - built-in function 'MPE-TRANSLATE-LOADED-OVERLAY-ADDRESS' */
xlValue xmpetranslateloadedoverlayaddress(void)
{
    MScalar addr;
    MPEDebugInfo *mpe;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); addr = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* translate the address */
    if (!mpeTranslateLoadedOverlayAddress(mpe,addr,&addr))
        return xlNil;

    /* return the translated address */
    return xlMakeFixnum((xlFIXTYPE)addr);
}

/* xmpepacketinfo - built-in function 'MPE-PACKET-INFO' */
void xmpepacketinfo(void)
{
    int length,type,cc;
    MPEDebugInfo *mpe;
    MScalar pc,ea;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); pc = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* get the packet information */
    if (!mpePacketInfo(mpe,pc,&length,&type,&cc,&ea)) {
        xlCPush(xlNil);
        xlMVReturn(1);
    }

    /* return the packet length and call indicator */
    xlCheck(4);
    xlPush(xlMakeFixnum((xlFIXTYPE)ea));
    xlPush(cc >= 0 ? xlMakeFixnum((xlFIXTYPE)cc) : xlFalse);
    xlPush(type ? xlMakeFixnum((xlFIXTYPE)type) : xlFalse);
    xlPush(xlMakeFixnum((xlFIXTYPE)length));
    xlMVReturn(4);
}

/* xmpeinitinstructiondecoder - built-in function 'MPE-INIT-INSTRUCTION-DECODER' */
xlValue xmpeinitinstructiondecoder(void)
{
    MPEDebugInfo *mpe;
    mpe = getMPE();
    xlLastArg();
    mpeFlushInstructionCache(mpe,64); /* IFetchCacheSize from mdlocal.h */
    return xlNil;
}

/* xmpedisassembleinstruction - built-in function 'MPE-DISASSEMBLE-INSTRUCTION' */
void xmpedisassembleinstruction(void)
{
    int length,endOfPacketP;
    char buf[256],*start,*p;
    MPEDebugInfo *mpe;
    xlValue last,next;
    MScalar pc,ea;
    int ch;

    /* parse the argument list */
    mpe = getMPE();
    xlVal = xlGetArgFixnum(); pc = (MScalar)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); ea = (MScalar)xlGetFixnum(xlVal);
    xlLastArg();

    /* disassemble the instruction */
    if ((length = mpeDisassembleInstruction(mpe,pc,ea,buf,sizeof(buf),&endOfPacketP)) == 0) {
        xlCPush(xlNil);
        xlMVReturn(1);
    }

    /* build a list of the instructions (may include register unit dec instructions) */
    start = p = buf;
    xlVal = xlNil;
    do {
        if ((ch = *p++) == '\n' || ch == '\0') {
            next = xlCons(xlMakeString(start,p - start - 1),xlNil);
            if (xlVal == xlNil)
                xlVal = next;
            else
                xlSetCdr(last,next);
            last = next;
            start = p;
        }
    } while (ch != '\0');

    /* return the instruction list, packet length and call indicator */
    xlCheck(3);
    xlPush(endOfPacketP ? xlTrue : xlFalse);
    xlPush(xlMakeFixnum((xlFIXTYPE)length));
    xlPush(xlVal);
    xlMVReturn(3);
}

/* xdisplaylines - display lines from a file */
xlValue xdisplaylines(void)
{
    xlFIXTYPE start,count;
    char buf[256],*p;
    xlValue name;
    FILE *fp;
    
    /* parse the argument list */
    name = xlGetArgString();
    xlVal = xlGetArgFixnum(); start = xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); count = xlGetFixnum(xlVal);
    xlVal = xlMoreArgsP() ? xlGetOutputPort() : xlCurOutput();
    xlLastArg();

    /* open the file */
    if ((fp = fopen(xlGetString(name),"r")) == NULL)
        return xlFalse;

    /* seek to the specified start line */
    while (fgets(buf,sizeof(buf),fp)) {

        /* look for the start line */
        if (start > 1)
            --start;

        /* display the specified number of lines */
        else if (count > 0) {
            for (p = buf; *p != '\0'; )
                xlPutC(xlVal,*p++);
            --count;
        }

        /* break when done */
        else
            break;
    }
    fclose(fp);

    /* return successfully */
    return count == 0 ? xlTrue : xlFalse;
}

/* FetchImageLines context structure */
typedef struct {
    xlValue fcn;
    xlFIXTYPE offset;
    xlFIXTYPE maxOffset;
    xlFIXTYPE y;
} FetchImageLinesData;

/* AppendPixelToRow - append the next pixel to the current row */
static void AppendPixelToRow(void *data,int x,int y,int yy,int cr,int cb)
{
    FetchImageLinesData *c = (FetchImageLinesData *)data;
    Color color;
    xlValue val;
    char *p;

    /* allocate a new string if necessary */
    if (xlTop() == xlNil) {
        xlSetTop(xlNewString(c->maxOffset));
        c->offset = 0;
    }

    /* convert yCrCb to rgb */
    MapYCrCbToRGB(yy,cr,cb,&color);

    /* store the next pixel */
    p = xlGetString(xlTop()) + c->offset;
    *p++ = color.r;
    *p++ = color.g;
    *p = color.b;

    /* check for the end of a row */
    if ((c->offset += 3) >= c->maxOffset) {
        xlCallFunction(&val,1,c->fcn,2,xlMakeFixnum(c->y),xlTop());
        xlSetTop(xlNil);
        ++c->y;
    }
}

/* xMapOverImageRows - built-in function 'map-over-image-rows' */
xlValue xMapOverImageRows(void)
{
    FetchImageLinesData c;
    MScalar width,height;
    MMPDebugInfo *mmp;
    int sts;

    /* parse the argument list */
    mmp = getMMP();
    xlVal = xlGetArg();
    xlLastArg();

    /* initialize the context structure */
    c.fcn = xlVal;
    mmpDisplayImageSize(mmp,&width,&height);
    c.maxOffset = width * 3;
    c.offset = 0;
    c.y = 0;

    /* initialize the current line */
    xlCPush(xlNil);

    /* display the image */
    sts = mmpDisplayImage(mmp,AppendPixelToRow,&c);

    /* pop the current line */
    xlDrop(1);
    
    /* return status */
    return sts ? xlTrue : xlFalse;
}

/* xYCrCbToRGB - map a YCrCb color to RGB */
static xlValue xYCrCbToRGB(void)
{
    Color color;
    int y,cr,cb;
    xlFIXTYPE n;

    /* parse the argument list */
    xlVal = xlGetArgFixnum(); n = xlGetFixnum(xlVal);
    if (xlMoreArgsP()) {
        y = (int)n;
        xlVal = xlGetArgFixnum(); cr = (int)xlGetFixnum(xlVal);
        xlVal = xlGetArgFixnum(); cb = (int)xlGetFixnum(xlVal);
    }
    else {
        y =  (int)((n >> 24) & 0xff);
        cr = (int)((n >> 16) & 0xff);
        cb = (int)((n >>  8) & 0xff);
    }
    xlLastArg();

    /* map the YCrCb color to RGB */
    MapYCrCbToRGB(y,cr,cb,&color);
    return xlMakeFixnum((xlFIXTYPE)((color.r << 24)
                                  | (color.g << 16)
                                  | (color.b << 8)));
}

/* xwriteimagetofile - display an image in an output window */
xlValue xwriteimagetofile(void)
{
    MScalar base,width,height,x,y,sts;
    int mode = 4;
    Image image;
    Color color;
    char *name;
    MMPDebugInfo *mmp;

    /* parse the argument list */
    xlVal = xlGetArgString(); name = xlGetString(xlVal);
    mmp = getMMP();
    xlVal = xlGetArgFixnum(); base = (int)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); width = (int)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); height = (int)xlGetFixnum(xlVal);
    if (xlMoreArgsP()) {
        xlVal = xlGetArgFixnum();
        mode = (int)xlGetFixnum(xlVal);
    }
    xlLastArg();

    /* initialize the image */
    if (imInitImage(&image,width,height) != IM_OKAY)
        return xlFalse;
    
    /* write each pixel to the image */
    for (y = 0; y < height; ++y) {
        for (x = 0; x < width; ++x) {
            MScalar pixel;
            mmpFetchScalar(mmp,base,&pixel); base += sizeof(MScalar);
            if (mode == 6) base += sizeof(MScalar);
            MapYCrCbToRGB((pixel >> 24) & 0xFF, /* y */
                          (pixel >> 16) & 0xFF, /* cr */
                          (pixel >> 8)  & 0xFF, /* cb */
                          &color);
            imSetPixel(&image,x,y,color);
        }
    }

    /* write the image to the output file */
    sts = imDumpImage(&image,name);
    
    /* free the image */
    imDestroyImage(&image);

    /* return with status */
    return sts == IM_OKAY ? xlTrue : xlFalse;
}
/* xwriterawimagetofile - write a raw YCrCb image, with no RGB conversion;
   Y goes in the G channel, Cr in R, and Cb in B
 */
xlValue xwriterawimagetofile(void)
{
    MScalar base,width,height,x,y,sts;
    int mode = 4;
    Image image;
    Color color;
    char *name;
    MMPDebugInfo *mmp;

    /* parse the argument list */
    xlVal = xlGetArgString(); name = xlGetString(xlVal);
    mmp = getMMP();
    xlVal = xlGetArgFixnum(); base = (int)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); width = (int)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); height = (int)xlGetFixnum(xlVal);
    if (xlMoreArgsP()) {
        xlVal = xlGetArgFixnum();
        mode = (int)xlGetFixnum(xlVal);
    }
    xlLastArg();

    /* initialize the image */
    if (imInitImage(&image,width,height) != IM_OKAY)
        return xlFalse;
    
    /* write each pixel to the image */
    for (y = 0; y < height; ++y) {
        for (x = 0; x < width; ++x) {
                        MScalar pixel;
            mmpFetchScalar(mmp,base,&pixel); base += sizeof(MScalar);
            if (mode == 6) base += sizeof(MScalar);
                        color.g = (pixel >> 24) & 0xff;  /* y */
                        color.r = (pixel >> 16) & 0xff;  /* cr */
                        color.b = (pixel >> 8) & 0xff;   /* cb */

                        imSetPixel(&image,x,y,color);
                }
    }

    /* write the image to the output file */
    sts = imDumpImage(&image,name);
    
    /* free the image */
    imDestroyImage(&image);

    /* return with status */
    return sts == IM_OKAY ? xlTrue : xlFalse;
}

/* xgetcwd - built-in function 'getcwd' */
static xlValue xgetcwd(void)
{
    char *str;
    xlLastArg();
#ifdef WIN32
    if ((str = _getcwd(NULL,0)) == NULL)
#else
    if ((str = getcwd(NULL,0)) == NULL)
#endif
        return xlNil;
    xlVal = xlMakeCString(str);
    free(str);
    return xlVal;
}

/* xconvertslashes - built-in function 'convert-slashes' */
static xlValue xconvertslashes(void)
{
    char *src,*dst;
    xlFIXTYPE len;
    xlValue str;
    int ch;

    /* parse the arguments */
    xlVal = xlGetArgString();
    xlLastArg();

    /* make a new string */
    len = xlGetSLength(xlVal);
    str = xlNewString(len);

    /* copy the string converting backslashes */
    src = xlGetString(xlVal);
    dst = xlGetString(str);
    while (--len >= 0)
        *dst++ = (ch = *src++) == '\\' ? '/' : ch;

    /* return the converted string */
    return str;
}

/* xreadmerlinsymbol - read a merlin symbol */
xlValue xreadmerlinsymbol(void)
{
    extern xlValue xlSymConst,xlSymNMacro;
    char buf[256];
    xlValue type;
    int ch,i;

    /* parse the arguments */
    xlVal = xlGetArgInputPort();
    xlLastArg();

    /* get symbol name */
    for (i = 0; (ch = xlGetC(xlVal)) != EOF && (type = xlCharType(ch)) == xlSymConst || type == xlSymNMacro; )
        if (i < xlSTRMAX)
            buf[i++] = ch;
    buf[i] = '\0';

    /* save the break character */
    xlUngetC(xlVal,ch);
    return buf[0] ? xlMakeCString(buf) : xlNil;
}

/* MComInitialize - enter merlin specific symbols and functions */
static void MComInitialize(void)
{
    /* enter the foreign pointer type */
    s_connection = xlEnter("MCOM-CONNECTION");

    /* enter the connection class */
    c_connection = xlClass("MCOM-CONNECTION",NULL,"HANDLE");
    xlMethod(c_connection,"INITIALIZE",xMComInitialize);
    xlMethod(c_connection,"CLOSE",xMComClose);
    xlMethod(c_connection,"DATA-AVAILABLE?",xMComDataAvailableP);
    xlXMethod(c_connection,"READ-PACKET",xMComReadPacket);
    xlMethod(c_connection,"WRITE-PACKET",xMComWritePacket);
}

/* xMComInitialize - built-in method 'initialize' */
static xlValue xMComInitialize(void)
{
    char *port = NULL;
    CommConnection *c;
    xlValue obj;
    long addr;

    /* parse the arguments */
    obj = xlGetArgInstance(c_connection);
    xlVal = xlGetArgFixnum(); addr = (long)xlGetFixnum(xlVal);
    if (xlMoreArgsP()) {
        xlVal = xlGetArgFixnum();
        addr = afVirtualAddr(addr,(long)xlGetFixnum(xlVal));
        if (xlMoreArgsP())
            port = xlGetString(xlGetArgString());
    }
    xlLastArg();

    /* default the port to the value of the environment variable MD_PORT */
    if (!port)
        port = getenv("MD_PORT");

    /* try to open the connection */
    if (!port || !(c = mcomOpen(port,addr)))
        return xlNil;

    /* return the connection */
    SetConnection(obj,c);
    return obj;
}

/* xMComClose - built-in function 'mcom-close' */
static xlValue xMComClose(void)
{
    CommConnection *c;

    /* parse the arguments */
    xlVal = xlGetArgInstance(c_connection);
    xlLastArg();

    /* make sure it's not already closed */
    if ((c = (CommConnection *)xlGetFPtr(xlVal)) == 0)
        xlError("attempt to close an already closed mcom connection",xlVal);

    /* close the connection */
    SetConnection(xlVal,0);
    mcomClose(c);

    /* return successfully */
    return xlTrue;
}

/* xMComDataAvailableP - built-in function 'mcom-data-available?' */
static xlValue xMComDataAvailableP(void)
{
    CommConnection *c;
    int timeout = 0;

    /* parse the arguments */
    c = GetConnection();
    if (xlMoreArgsP()) {
        xlVal = xlGetArgFixnum(); timeout = (int)xlGetFixnum(xlVal);
    }
    xlLastArg();

    /* check to see if data is available */
    return mcomReadDataAvailableP(c,timeout) ? xlTrue : xlFalse;
}

/* xMComReadPacket - built-in function 'mcom-read-packet' */
static void xMComReadPacket(void)
{
    int timeout,timeoutP = FALSE;
    CommConnection *c;
    long v0,v1,v2,v3;

    /* parse the arguments */
    c = GetConnection();
    if (xlMoreArgsP()) {
        xlVal = xlGetArgFixnum(); timeout = (int)xlGetFixnum(xlVal);
        timeoutP = TRUE;
    }
    xlLastArg();

    /* make sure data is available if a timeout was requested */
    if (timeoutP) {
        if (!mcomReadDataAvailableP(c,timeout))
            xlMVReturn(0);
    }

    /* read the packet */
    if (!mcomReadPacket(c,&v0,&v1,&v2,&v3))
        xlMVReturn(0);

    /* return the packet */
    xlCheck(4);
    xlPush(xlMakeFixnum((xlFIXTYPE)v3));
    xlPush(xlMakeFixnum((xlFIXTYPE)v2));
    xlPush(xlMakeFixnum((xlFIXTYPE)v1));
    xlPush(xlMakeFixnum((xlFIXTYPE)v0));
    xlMVReturn(4);
}

/* xMComWritePacket - built-in function 'mcom-write-packet' */
static xlValue xMComWritePacket(void)
{
    CommConnection *c;
    long v0,v1,v2,v3;

    /* parse the arguments */
    c = GetConnection();
    xlVal = xlGetArgFixnum(); v0 = (long)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); v1 = (long)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); v2 = (long)xlGetFixnum(xlVal);
    xlVal = xlGetArgFixnum(); v3 = (long)xlGetFixnum(xlVal);
    xlLastArg();

    /* write the packet */
    return mcomWritePacket(c,v0,v1,v2,v3) ? xlTrue : xlFalse;
}

/* GetConnection - get a mmp memory argument */
static CommConnection *GetConnection(void)
{
    xlValue arg,handle;
    CommConnection *c;

    /* get the next argument */
    arg = xlGetArgInstance(c_connection);

    /* get the connection handle */
    handle = xlGetIVar(arg,connectionHANDLE);
    if (!xlForeignPtrP(handle) || xlGetFPType(handle) != s_connection)
        xlError("bad handle in connection object",handle);

    /* make sure it's not closed */
    if ((c = (CommConnection *)xlGetFPtr(handle)) == 0)
        xlError("attempt to use a closed mcom connection",handle);

    /* return the connection */
    return (CommConnection *)xlGetFPtr(handle);
}

/* SetConnection - set the widget structure of a widget object */
static void SetConnection(xlValue obj,CommConnection *c)
{
    xlSetIVar(obj,connectionHANDLE,xlMakeForeignPtr(s_connection,c));
}


#ifdef WIN32
#define DIRSEP '\\'
#define ENVSEP ';'
#else
#define DIRSEP '/'
#define ENVSEP ':'
#endif

static char *GetProgramDirectory(char *buf,int size);
static char *GetWorkingDirectory(char *buf,int size);

/* LoadPath - get the load path for XLisp files

  Search order is 1) current working dir
                  2) path set by PUFFIN_PATH env var
                     or <exe-path>/../user;
                        <exe-path>/../util/puffin;
                        <exe-path>/../util/xlisp
*/
char *LoadPath(void)
{
#ifdef WIN32
    char *stdDirs[] = { "user",
                        "util\\puffin",
                        "util\\xlisp",
                        NULL };
#else
    char *stdDirs[] = { "user",
                        "util/puffin",
                        "util/xlisp",
                        NULL };
#endif
    static char xlispLoadPath[MAX_LOADPATH_LEN];
    char *end,*next,*pch;
    int i;
    
    /* initialize */
    end = xlispLoadPath + MAX_LOADPATH_LEN - 1;
    next = xlispLoadPath;

    /* get the current working directory */
    if (!GetWorkingDirectory(next,end - next))
        xlispLoadPath[0] = '\0';
    next += strlen(next);

    /* get the vmlabs directory first using the VMLABS environment variable */
    if ((pch = getenv("VMLABS")) != NULL)
        strncpy(vmlabsPath,pch,MAX_LOADPATH_LEN);
    
    /* try using the directory where the executable was found */
    else if (GetProgramDirectory(vmlabsPath,MAX_LOADPATH_LEN)) {

        /* remove \bin from the stdPath */
        if ((pch = strrchr(vmlabsPath,DIRSEP)) != NULL)
            *pch = '\0';
    }

    /* default to the current directory */
    else
        strcpy(vmlabsPath,".");

    /* first try using PUFFIN_PATH for the load path */
    if ((pch = getenv("PUFFIN_PATH")) != NULL)  {
        if (xlispLoadPath[0] && next < end)
            *next++ = ENVSEP;
        strncat(next,pch,end - next);
    }

    /* if that fails, use the vmlabs directory */
    else {
        
        /* append all of the standard xlisp search paths to 
           whatever may have been in PUFFIN_PATH */
        for (i = 0; stdDirs[i] != NULL; i++) {

            /* add a separator if appending to existing path */
            if (xlispLoadPath[0] && next < end)
                *next++ = ENVSEP;

            /* append the directory prefix */
            if (vmlabsPath[0]) {
                strncat(next,vmlabsPath,end - next);
                next += strlen(next);
                if (next < end)
                    *next++ = DIRSEP;
            }

            /* append the subdirectory */
            strncat(next,stdDirs[i],end - next);
            next += strlen(next);
        }
    }

    /* return the load path */
    return xlispLoadPath;
}

/* GetProgramDirectory - get the directory containing the program */
static char *GetProgramDirectory(char *buf,int size)
{
#ifdef WIN32
    long pathLen;
    char *pch;

    /* get the executable location */
    pathLen = SearchPath(NULL,"xlisp.dll",NULL,size,buf,&pch);

    /* check to see if the call succeeded */
    if (pathLen == 0)
        return NULL;

    /* remove exe name from the path */
    *--pch = '\0';

    /* return the buffer */
    return buf;
#else
    return NULL;
#endif
}

/* GetWorkingDirectory - get the working directory */
static char *GetWorkingDirectory(char *buf,int size)
{
#ifdef WIN32
    return _getcwd( buf, size );
#else
    return getcwd( buf, size );
#endif
}

#ifdef WIN32

static int
osExecuteArgv(char **argv)
{
    STARTUPINFO startInfo;
    PROCESS_INFORMATION procInfo;
    DWORD exitSts;
    HANDLE in,out,err;
    BOOL sts;
    char cmdLine[512];
    int i;

    if (!argv[0])
        return -1;
    strcpy(cmdLine,argv[0]);
    for (i = 1; argv[i] ; i++) {
        strcat(cmdLine," ");
        strcat(cmdLine,argv[i]);
    }

    memset(&startInfo,0,sizeof(startInfo));
    startInfo.cb = sizeof(startInfo);
    startInfo.dwFlags = STARTF_USESTDHANDLES;
    if ((in = CreateFile("NUL",
                         GENERIC_READ,
                         0,
                         NULL,
                         OPEN_EXISTING,
                         FILE_ATTRIBUTE_NORMAL,
                         NULL)) == INVALID_HANDLE_VALUE)
        return -1;
    if ((out = CreateFile("NUL",
                          GENERIC_WRITE,
                          0,
                          NULL,
                          OPEN_EXISTING,
                          FILE_ATTRIBUTE_NORMAL,
                          NULL)) == INVALID_HANDLE_VALUE)
        return -1;
    if ((err = CreateFile("NUL",
                          GENERIC_WRITE,
                          0,
                          NULL,
                          OPEN_EXISTING,
                          FILE_ATTRIBUTE_NORMAL,
                          NULL)) == INVALID_HANDLE_VALUE)
        return -1;
    startInfo.hStdInput = in;
    startInfo.hStdOutput = out;
    startInfo.hStdError = err;
    sts = CreateProcess(
            NULL,               /* application name */
            cmdLine,            /* command line */
            NULL,               /* security attributes */
            NULL,               /* thread attributes */
            FALSE,              /* inherit handles */
            DETACHED_PROCESS,   /* creation flags */
            NULL,               /* environment */
            NULL,               /* current directory */
            &startInfo,         /* startup info */
            &procInfo);         /* process info */
    if (!sts) return -1;
    WaitForSingleObject(procInfo.hProcess,(DWORD)-1);
    GetExitCodeProcess(procInfo.hProcess,&exitSts);
    return exitSts;
}

#endif

/* xrunprogram - run a program */
/* this one takes just a single string */
xlValue xrunprogram(void)
{
#ifdef WIN32
#define MAXSPAWN    32
    char *args[MAXSPAWN + 1];
    int count = 0;
    int status;
    
    while (xlMoreArgsP() && count < MAXSPAWN)
        args[count++] = xlGetString(xlGetArgString());
    xlLastArg();
    args[count] = NULL;
    status = osExecuteArgv(args);
    if (status < 0)
        return xlNil;
    return xlMakeFixnum((xlFIXTYPE)status);
#else
    int status;
    char cmdLine[512];

    strcpy(cmdLine,xlGetString(xlGetArgString()));
    while (xlMoreArgsP()) {
        strcat(cmdLine," ");
        strcat(cmdLine,xlGetString(xlGetArgString()));
    }
    xlLastArg();

    status = system(cmdLine);

    /* /bin/sh returns 127 for command not found */
    /* HOWEVER: the Linux system() function puts the status in
     * the wrong place... make sure it gets moved to the right one
     */
    if ( ((status & 0x00ff) == 0) && (status & 0xff00) != 0 )
        status = status >> 8;

    if (status >= 120)
        return xlNil;
    return xlMakeFixnum((xlFIXTYPE)status);
#endif
}

/* osCallHandler - call a handler function */
int osCallHandler(char *name)
{
    xlValue val;
    int vcnt;
    if (!xlInitializedP)
        return -1;
    vcnt = xlCallFunctionByName(&val,1,name,0);
    xlosFlushOutput();
    return vcnt >= 0 ? (xlFixnumP(val) ? (int)xlGetFixnum(val) : 0) : -1;
}

/* osCallHandlerWithNumericArg - call a handler function */
int osCallHandlerWithNumericArg(char *name,int arg)
{
    xlValue val;
    int vcnt;
    if (!xlInitializedP)
        return -1;
    vcnt = xlCallFunctionByName(&val,1,name,1,xlMakeFixnum(arg));
    xlosFlushOutput();
    return vcnt >= 0 ? (xlFixnumP(val) ? (int)xlGetFixnum(val) : 0) :  -1;
}

/* getMMP - get a mmp memory argument */
MMPDebugInfo *getMMP(void)
{
    xlValue arg = xlGetArgForeignPtr();
    MMPDebugInfo *mmp;
    if (xlGetFPType(arg) != s_MMP)
        xlError("expecting mmp memory",arg);
    else if ((mmp = (MMPDebugInfo *)xlGetFPtr(arg)) == 0)
        xlError("attempt to use deleted mmp memory",arg);
    return mmp;
}

/* getMPE - get a mpe memory argument */
MPEDebugInfo *getMPE(void)
{
    xlValue arg = xlGetArgForeignPtr();
    MPEDebugInfo *mpe;
    if (xlGetFPType(arg) != s_MPE)
        xlError("expecting mpe memory",arg);
    else if ((mpe = (MPEDebugInfo *)xlGetFPtr(arg)) == 0)
        xlError("attempt to use deleted mpe memory",arg);
    return mpe;
}

/* getTYPE - get a type argument */
static CoffType *getTYPE(void)
{
    xlValue arg = xlGetArgForeignPtr();
    CoffType *type;
    if (xlGetFPType(arg) != s_TYPE)
        xlError("expecting type",arg);
    else if ((type = (CoffType *)xlGetFPtr(arg)) == 0)
        xlError("attempt to use deleted type",arg);
    return type;
}
