/* db_vmavr.c - a simple virtual machine for the Chameleon AVR system
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include "CHAM_AVR_SYSTEM_V010.h"
#include "CHAM_AVR_TWI_SPI_DRV_V010.h"			

#include <stdio.h>
#include <util/delay.h> 
#include "../runtime/db_vm.h"
#include "uart.h"			

/* code space in flash */
uint8_t __attribute__((section(".vmimage"))) vmimage[VMCODE_SIZE] = {
#include "../runtime/ready.h"
};

/* data space in sram */
uint8_t vmdata[VMDATA_SIZE] VMDATA_ATTR;

void VM_sysinit(int argc, char *argv[], const uint8_t **image, uint8_t **data, uint16_t *dataSize)
{
    UART_init(57600);//115200);
    SPI_Init(SPI_DEFAULT_RATE);
    SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);
    _delay_ms(2000);
    *image = vmimage;
    *data = vmdata;
    *dataSize = VMDATA_SIZE;
}

int VM_getchar(void)
{
    int ch;
    while ((ch = UART_getc()) == -1)
        ;
    return ch;
}

void VM_putchar(int ch)
{
    if (ch == '\n')
        UART_putc('\r');
    UART_putc(ch);
}

void VM_flush(void)
{
}

void vmcodecopy(uint8_t *dst, const uint8_t *src, int cnt)
{
    while (--cnt >= 0)
        *dst++ = VMCODEBYTE(src++);
}

int vmstrlen(const uint8_t *str)
{
    int len = 0;
    while (VMCODEBYTE(str++) != '\0')
        ++len;
    return len;
}

char *itoa(int value, char *string, int radix)
{
    switch (radix) {
    case 8:
        sprintf(string, "%o", value);
        break;
    case 10:
        sprintf(string, "%d", value);
        break;
    case 16:
        sprintf(string, "%x", value);
        break;
    default:
        strcpy(string, "<unknown radix>");
        break;
    }
    return string;
}


