/* db_vmpic.c - a simple virtual machine for the XGS PIC system
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include "CHAM_PIC_SYSTEM_V010.h"
#include "CHAM_PIC_UART_DRV_V010.h"			
#include "CHAM_PIC_I2C_SPI_DRV_V010.h"			

#include <stdio.h>
#include "../runtime/db_vm.h"

/* code space in flash */
const uint8_t __attribute__((address(0x8000), space(prog))) vmimage[VMCODE_SIZE] = {
#include "../runtime/ready.h"
};

/* data space in sram */
uint8_t vmdata[VMDATA_SIZE] VMDATA_ATTR;

void VM_sysinit(int argc, char *argv[], const uint8_t **image, uint8_t **data, uint16_t *dataSize)
{
    SYS_ConfigureClock(MAX_FCY_RATE);
    UART_Init(57600);
	SPI_SET_CS(SPI_CS_0);
	SPI_Init();
    DELAY_MS(2000);
    TBLPAG = __builtin_tblpage(vmimage);
    *image = vmimage;
    *data = vmdata;
    *dataSize = VMDATA_SIZE;
}

int VM_getchar(void)
{
    unsigned char ch;
    while (!UART_getchar(&ch))
        ;
    return ch;
}

void VM_putchar(int ch)
{
    if (ch == '\n')
        UART_putchar('\r');
    UART_putchar(ch);
}

void VM_flush(void)
{
}

void vmcodecopy(uint8_t *dst, const uint8_t *src, int cnt)
{
    while (--cnt >= 0)
        *dst++ = VMCODEBYTE(src++);
}

int vmstrlen(const uint8_t *str)
{
    int len = 0;
    while (VMCODEBYTE(str++) != '\0')
        ++len;
    return len;
}

char *itoa(int value, char *string, int radix)
{
    switch (radix) {
    case 8:
        sprintf(string, "%o", value);
        break;
    case 10:
        sprintf(string, "%d", value);
        break;
    case 16:
        sprintf(string, "%x", value);
        break;
    default:
        strcpy(string, "<unknown radix>");
        break;
    }
    return string;
}
