/* db_types.h - type definitions for a simple virtual machine
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#ifndef __DB_TYPES_H__
#define __DB_TYPES_H__

/**********/
/* Common */
/**********/

#define VMTRUE		1
#define VMFALSE   	0

/*********/
/* WIN32 */
/*********/

#ifdef WIN32

#include "../common/db_inttypes.h"

#include <stdio.h>
#include <string.h>

int vsnprintf(char *str, size_t size, const char *format, va_list ap);

typedef FILE VMFILE;

#define VMCODE_SIZE		        (16 * 1024)
#define VMDATA_SIZE    	        11720
#define VMDATA_ATTR

#define FLASH_SPACE
#define FLASH_PAGESIZE          512
#define FLASH_BOOT

#define VMCODEBYTE(p)           *(uint8_t *)(p)
#define VMCODECOPY(d, s, c)     memcpy(d, s, c)
#define VMINTRINSIC(i)          Intrinsics[i]
#define VMCOPY(d, s, c, h)      memcpy(d, s, c)
#define VMGETBYTE(p, h)         *(uint8_t *)(p)
#define VMSTRLEN(p)             strlen(p)
#define VMSTRNCMP(a, b, n)      strncmp(a, b, n)
#define VMVSNPRINTF             vsnprintf

#define DEFAULT_HEAP_SIZE       1024
#define DEFAULT_HEAP_HANDLES    32

#endif  // WIN32

/**********/
/* XGSPIC */
/**********/

#ifdef XGSPIC

#include "FSIO.h"

typedef FSFILE VMFILE;

#define VMCODE_SIZE		        (16 * 1024)
#define VMDATA_SIZE    	        11600

#define DEFAULT_HEAP_SIZE       1024
#define DEFAULT_HEAP_HANDLES    32

#define XGS_COMMON
#define PIC_COMMON

#endif  // XGSPIC

/****************/
/* CHAMELEONPIC */
/****************/

#ifdef CHAMELEONPIC

typedef void VMFILE;

#define VMCODE_SIZE		        32767
#define VMDATA_SIZE    	        2048

#define DEFAULT_HEAP_SIZE       1024
#define DEFAULT_HEAP_HANDLES    32

#define PIC_COMMON

#endif  // CHAMELEONPIC

/**************/
/* PIC common */
/**************/

#ifdef PIC_COMMON

#include "db_inttypes.h"

#include <stdio.h>
#include <string.h>
#include <libpic30.h>
#include <p24hxxxx.h>

// not defined but actually located in the gcc library!
int vsnprintf(char *, size_t, const char *, va_list);

void vmcodecopy(uint8_t *dst, const uint8_t *src, int cnt);
int vmstrlen(const uint8_t *str);
char *itoa(int value, char *string, int radix);

#define VMDATA_ATTR             __attribute__((far))

#define FLASH_SPACE             const
#define FLASH_PAGESIZE          (_FLASH_PAGE * 2)
#define FLASH_ROWSIZE           (_FLASH_ROW * 2)
#define FLASH_BOOT

#define VMCODEBYTE(p)           ({                              \
                                    register int v;             \
                                    asm("tblrdl.b [%1], %0\n\t" \
                                        "ze %0, %0"             \
                                        : "=r"(v) : "r"(p));    \
                                    v;                          \
                                })                                
#define VMCODECOPY(d, s, c)     vmcodecopy(d, s, c)
#define VMSTRLEN(p)             vmstrlen(p)
#define VMINTRINSIC(i)          Intrinsics[i]
#define VMCOPY(d, s, c, h)      do {                            \
                                    if (h)                      \
                                        memcpy(d, s, c);        \
                                    else                        \
                                        VMCODECOPY(d, s, c);    \
                                } while (0)
#define VMGETBYTE(p, h)         ((h) ? *(uint8_t *)(p) : VMCODEBYTE(p))
#define VMSTRNCMP(a, b, n)      strncmp(a, b, n)
#define VMVSNPRINTF             vsnprintf

#endif  // PIC_COMMON

/**********/
/* XGSAVR */
/**********/

#ifdef XGSAVR

#define VMCODE_SIZE		        (4 * 1024)
#define VMDATA_SIZE    	        2200

#define DEFAULT_HEAP_SIZE       512
#define DEFAULT_HEAP_HANDLES    16

#define XGS_COMMON
#define AVR_COMMON

#endif  // XGSAVR

/****************/
/* CHAMELEONAVR */
/****************/

#ifdef CHAMELEONAVR

#define VMCODE_SIZE		        (2 * 1024)
#define VMDATA_SIZE    	        1300

#define DEFAULT_HEAP_SIZE       256
#define DEFAULT_HEAP_HANDLES    8

#define AVR_COMMON

#endif  // XGSAVR

/**********/
/* AVR328 */
/**********/

#ifdef AVR328

#define VMCODE_SIZE		        (4 * 1024)
#define VMDATA_SIZE    	        1024

#define DEFAULT_HEAP_SIZE       256
#define DEFAULT_HEAP_HANDLES    8

#define AVR_COMMON

#endif  // AVR328

/**********/
/* UZEBOX */
/**********/

#ifdef UZEBOX

#define VMCODE_SIZE		        (4 * 1024)
#define VMDATA_SIZE    	        512

#define DEFAULT_HEAP_SIZE       256
#define DEFAULT_HEAP_HANDLES    8

#define AVR_COMMON

#endif  // UZEBOX

/**************/
/* AVR common */
/**************/

#ifdef AVR_COMMON

#include <stdint.h>
#include <string.h>
#include <avr/boot.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

typedef void VMFILE;

#define VMDATA_ATTR

#define FLASH_SPACE             PROGMEM
#define FLASH_PAGESIZE          SPM_PAGESIZE
#define FLASH_BOOT              BOOTLOADER_SECTION

#define VMCODEBYTE(p)	        pgm_read_byte(p)
#define VMCODECOPY(d, s, c)     memcpy_P(d, s, c)
#define VMINTRINSIC(i)          ((IntrinsicFcn *)pgm_read_word(&Intrinsics[i]))
#define VMCOPY(d, s, c, h)      do {                        \
                                    if (h)                  \
                                        memcpy(d, s, c);    \
                                    else                    \
                                        memcpy_P(d, s, c);  \
                                } while (0)
#define VMGETBYTE(p, h)         ((h) ? *(uint8_t *)(p) : VMCODEBYTE(p))
#define VMSTRLEN(p)             strlen_P((char *)p)
#define VMSTRNCMP(a, b, n)      strncmp_P((char *)a, (char *)b, n)
#define VMVSNPRINTF             vsnprintf_P

#endif  // AVR

#endif
